/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2001, 2006 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* NXAGENT, NX protocol compression and NX extensions to this software    */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for applicability.       */
/*                                                                        */
/* NX and NoMachine are trademarks of Medialogic S.p.A.                   */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

/*

Copyright 1993 by Davor Matic

Permission to use, copy, modify, distribute, and sell this software
and its documentation for any purpose is hereby granted without fee,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation.  Davor Matic makes no representations about
the suitability of this software for any purpose.  It is provided "as
is" without express or implied warranty.

*/

#include <stdio.h>
#include <string.h>
#include <errno.h>

#ifdef __sun
#include <strings.h>
#endif

#include "X.h"
#include "Xproto.h"
#include "screenint.h"
#include "input.h"
#include "misc.h"
#include "scrnintstr.h"
#include "servermd.h"
#include "opaque.h"

#include "Agent.h"
#include "Display.h"
#include "Args.h"
#include "Options.h"
#include "Binder.h"
#include "Trap.h"
#include "Screen.h"
#include "Image.h"
#ifdef RENDER
#include "Render.h"
#endif

/*
 * NX specific includes and definitions.
 */

#include "NXlib.h"
#include "NXpack.h"

/*
 * Set here the required log level.
 */

#define PANIC
#define WARNING
#undef  TEST
#undef  DEBUG
#undef  WATCH

#ifdef WATCH

#include "unistd.h"

#endif

int nxagentUserDefinedFontPath = 0;

/*
 * From X11/ImUtil.c.
 */

extern int _XGetBitsPerPixel(Display *dpy, int depth);

/*
 * Pack and split parameters gotten we
 * get from the NX transport.
 */

int nxagentPackLossless   = -1;
int nxagentPackMethod     = -1;
int nxagentPackQuality    = -1;
int nxagentSplitThreshold = -1;
int nxagentTightThreshold = -1;

extern char dispatchExceptionAtReset;

char nxagentDisplayName[1024];
char nxagentWindowName[256];
char nxagentDialogName[256];
char nxagentSessionId[256];
char *nxagentOptionFile;

Bool nxagentFullGeneration = False;
int nxagentDefaultClass = TrueColor;
Bool nxagentUserDefaultClass = False;
int nxagentDefaultDepth;
Bool nxagentUserDefaultDepth = False;
struct UserGeometry nxagentUserGeometry = {0, 0, 0, 0, 0};
Bool nxagentUserBorderWidth = False;
int nxagentNumScreens = 0;
Bool nxagentDoDirectColormaps = False;
Window nxagentParentWindow = 0;
Bool nxagentIpaq = False;

Bool nxagentResizeDesktopAtStartup = False;

Bool nxagentUseNXTrans   = False;
Bool nxagentForceNXTrans = False;

int nxagentMaxAllowedReset = 0;

char *nxagentKeyboard = NULL;

#ifdef NXAGENT_ONEXIT
Bool nxagentOnce = True;
#endif

static void nxagentParseOptionString(char*);

/*
 * Get the caption to be used for helper dialogs
 * from agent's window name passed as parameter.
 */

static int nxagentGetDialogName(void);

int ddxProcessArgument(int argc, char *argv[], int i)
{
  /*
   * Ensure that the options are set to their defaults.
   */

  static Bool resetOptions = True;

  if (resetOptions == True)
  {
    char *envOptions = NULL;
    char *envDisplay = NULL;
    int j;

    nxagentResetOptions();

    resetOptions = False;

   /*
    * All the following lines before closed bracket
    * have been added to ensure the correct order of
    * options evaluation; first the environment ones,
    * then those included in a option files, and last
    * the remaining command line options.
    */

    envDisplay = getenv("DISPLAY");

    if (envDisplay != NULL && strlen(envDisplay) == 0)
    {
      envDisplay = NULL;
    }

    for (j = 0; j < argc; j++)
    {
      if ((!strcmp(argv[j], "-display")) && (j + 1 < argc))
      {
        envOptions = malloc(strlen(argv[j + 1]) + 1);

        if (envOptions != NULL)
        {
          envOptions = strcpy(envOptions, argv[j + 1]);
        }
        #ifdef WARNING
        else
        {
          fprintf(stderr, "ddxProcessArgument: WARNING! failed string allocation.\n");
        }
        #endif

        break;
      }
    }

    if ((envOptions == NULL) && (envDisplay != NULL))
    {
      envOptions = malloc(strlen(envDisplay) + 1);

      if (envOptions != NULL)
      {
        envOptions = strcpy(envOptions, envDisplay);
      }
      #ifdef WARNING
      else
      {
        fprintf(stderr, "ddxProcessArgument: WARNING! failed string allocation.\n");
      }
      #endif
    }

    if (envOptions != NULL)
    {
      nxagentParseOptionString(envOptions);
      free(envOptions);
    }

    for (j = 0; j < argc; j++)
    {
      if ((!strcmp(argv[j], "-options") || !strcmp(argv[j], "-option")) && j + 1 < argc)
      {
	if (nxagentOptionFile)
        {
          nxagentOptionFile = (char *) realloc(nxagentOptionFile, strlen(argv[j + 1]) + 1);
        }
        else
        {
          nxagentOptionFile = (char *) malloc(strlen(argv[j + 1]) +1);
        }

        if (nxagentOptionFile != NULL)
        {
          nxagentOptionFile = strcpy(nxagentOptionFile, argv[j + 1]);
        }
        #ifdef WARNING
        else
        {
          fprintf(stderr, "ddxProcessArgument: WARNING! failed string allocation.\n");
        }
        #endif

        break;
      }
    }

   if (nxagentOptionFile)
   {
     nxagentProcessOptionsFile();
   }

  }

  if (!strcmp(argv[i], "-B"))
  {
    #ifdef TEST
    fprintf(stderr, "ddxProcessArgument: Checking the NX binder option.\n");
    #endif

    if (nxagentCheckBinder(argc, argv, i) > 0)
    {
      /*
       * We are going to run the agent with the
       * 'binder' option. Go straight to the
       * proxy loop.
       */

      #ifdef TEST
      fprintf(stderr, "ddxProcessArgument: Entering the NX proxy binder loop.\n");
      #endif

      nxagentBinderLoop();

      /*
       * This will make an exit.
       */

      nxagentBinderExit(0);
    }
    else
    {
      /*
       * Exit with an error.
       */

      nxagentBinderExit(1);
    }
  }

  if (!strcmp(argv[i], "-display"))
  {
    if (++i < argc)
    {
      strncpy(nxagentDisplayName, argv[i], 1023);

      nxagentDisplayName[1023] = '\0';

      return 2;
    }

    return 0;
  }

  if (!strcmp(argv[i], "-id"))
  {
    if (++i < argc)
    {
      strncpy(nxagentSessionId, argv[i], 255);

      *(nxagentSessionId + 255) = '\0';

      return 2;
    }

    return 0;
  }

  /*
   * This had to be '-options' since the beginning
   * but was '-option' by mistake. Now we have to
   * handle the backward compatibility.
   */

  if (!strcmp(argv[i], "-options") || !strcmp(argv[i], "-option"))
  {
    if (++i < argc)
    {
      int size;

      if (nxagentOptionFile != NULL)
      {
        xfree(nxagentOptionFile);

        nxagentOptionFile = NULL;
      }

      if ((size = strlen(argv[i])) < 1024)
      {
        if ((nxagentOptionFile = xalloc(size + 1)) == NULL)
        {
          FatalError("malloc failed");
        }

        strncpy(nxagentOptionFile, argv[i], size);

        nxagentOptionFile[size] = '\0';
      }
      else
      {
        /*
         * It is useless to store the file name
         * that has just been truncated.
         */

        #ifdef WARNING
        fprintf(stderr, "ddxProcessArgument: WARNING! Option file name "
                    "too long. It will be ignored.\n");
        #endif
      }

      return 2;
    }

    return 0;
  }

  if (!strcmp(argv[i], "-full")) {
    nxagentFullGeneration = True;
    return 1;
  }

  if (!strcmp(argv[i], "-class")) {
    if (++i < argc) {
      if (!strcmp(argv[i], "StaticGray")) {
        nxagentDefaultClass = StaticGray;
        nxagentUserDefaultClass = True;
        return 2;
      }
      else if (!strcmp(argv[i], "GrayScale")) {
        nxagentDefaultClass = GrayScale;
        nxagentUserDefaultClass = True;
        return 2;
      }
      else if (!strcmp(argv[i], "StaticColor")) {
        nxagentDefaultClass = StaticColor;
        nxagentUserDefaultClass = True;
        return 2;
      }
      else if (!strcmp(argv[i], "PseudoColor")) {
        nxagentDefaultClass = PseudoColor;
        nxagentUserDefaultClass = True;
        return 2;
      }
      else if (!strcmp(argv[i], "TrueColor")) {
        nxagentDefaultClass = TrueColor;
        nxagentUserDefaultClass = True;
        return 2;
      }
      else if (!strcmp(argv[i], "DirectColor")) {
        nxagentDefaultClass = DirectColor;
        nxagentUserDefaultClass = True;
        return 2;
      }
    }
    return 0;
  }

  if (!strcmp(argv[i], "-cc")) {
    if (++i < argc && sscanf(argv[i], "%i", &nxagentDefaultClass) == 1) {
      if (nxagentDefaultClass >= 0 && nxagentDefaultClass <= 5) {
        nxagentUserDefaultClass = True;
        /* lex the OS layer process it as well, so return 0 */
      }
    }
    return 0;
  }

  if (!strcmp(argv[i], "-depth")) {
    if (++i < argc && sscanf(argv[i], "%i", &nxagentDefaultDepth) == 1) {
      if (nxagentDefaultDepth > 0) {
        nxagentUserDefaultDepth = True;
        return 2;
      }
    }
    return 0;
  }

  /*
   * These options are now useless and are
   * parsed only for compatibility with
   * older versions.
   */

  if (!strcmp(argv[i], "-fast") ||
          !strcmp(argv[i], "-slow") ||
              !strcmp(argv[i], "-hint") ||
                  !strcmp(argv[i], "-sss"))
  {
    fprintf(stderr, "Warning: Ignoring deprecated command line option '%s'.\n",
                argv[i]);

    return 1;
  }

  if (!strcmp(argv[i], "-backingstore"))
  {
    if (++i < argc)
    {
      if (!strcmp(argv[i], "0"))
      {
        nxagentChangeOption(BackingStore, BackingStoreNever);
      }
      else
      {
        nxagentChangeOption(BackingStore, BackingStoreForce);
      }

      return 2;
    }

    return 0;
  } 

  if (!strcmp(argv[i], "-streaming"))
  {
    if (++i < argc)
    {
      if (!strcmp(argv[i], "0"))
      {
        nxagentChangeOption(Streaming, 0);

/*
FIXME: As the streaming is temporarily disabled,
       we use its option value to set the Lazy.
*/
        nxagentChangeOption(DefaultLazyLevel, 0);
        nxagentChangeOption(LazyLevel, -1);
      }
      else
      {
        nxagentChangeOption(Streaming, 1);

/*
FIXME: Read the comment above.
*/
        nxagentSetDefaultLazyLevel();
      }

      return 2;
    }

    return 0;
  }

  if (!strcmp(argv[i], "-lazylevel"))
  {
    int level;

    if (++i < argc && sscanf(argv[i], "%i", &level) == 1
           && level >= 0 && level <= 4)
    {
      nxagentChangeOption(LazyLevel, level);

      return 2;
    }

    return 0;
  }

  if (strcmp(argv[i], "-fp") == 0)
  {
    if(++i < argc)
    {

#ifdef sgi
      userdefinedfontpath = 1;
#endif

      #ifdef TEST
      fprintf(stderr, "ddxProcessArgument: User defined font path [%s].\n", argv[i]);
      #endif

      nxagentUserDefinedFontPath = 1;

      defaultFontPath = argv[i];
    }
    else
    {
      UseMsg();
    }
  }

  if (!strcmp(argv[i], "-geometry"))
  {
    if (++i < argc)
    {
      if (!strcmp(argv[i],"fullscreen"))
      {
        nxagentChangeOption(Fullscreen, True);
      }
      else if (!strcmp(argv[i],"ipaq"))
      {
        nxagentChangeOption(Fullscreen, True);

        nxagentIpaq = True;
      }
      else
      { 
        if (nxagentUserGeometry.flag == 0) 
        {
          nxagentUserGeometry.flag = XParseGeometry(argv[i], 
                                                        &nxagentUserGeometry.X, 
                                                            &nxagentUserGeometry.Y, 
                                                                &nxagentUserGeometry.Width, 
                                                                    &nxagentUserGeometry.Height);
        }
      }

      if (nxagentUserGeometry.flag || (nxagentOption(Fullscreen) == 1)) return 2;
    }

    return 0;
  }

  if (!strcmp(argv[i], "-bw"))
  {
    int BorderWidth;

    if (++i < argc && sscanf(argv[i], "%i", &BorderWidth) == 1)
    {
      nxagentChangeOption(BorderWidth, BorderWidth);

      if (nxagentOption(BorderWidth) >= 0)
      {
        nxagentUserBorderWidth = True;

        return 2;
      }
    }

    return 0;
  }

  if (!strcmp(argv[i], "-name"))
  {
    if (++i < argc)
    {
      strncpy(nxagentWindowName, argv[i], 255);

      *(nxagentWindowName + 255) = '\0';

      return 2;
    }

    return 0;
  }

  if (!strcmp(argv[i], "-scrns")) {
    if (++i < argc && sscanf(argv[i], "%i", &nxagentNumScreens) == 1) {
      if (nxagentNumScreens > 0) {
        if (nxagentNumScreens > MAXSCREENS) {
          ErrorF("Maximum number of screens is %d.\n", MAXSCREENS);
          nxagentNumScreens = MAXSCREENS;
        }
        return 2;
      }
    }
    return 0;
  }

  if (!strcmp(argv[i], "-install")) {
    nxagentDoDirectColormaps = True;
    return 1;
  }

  if (!strcmp(argv[i], "-parent")) {
    if (++i < argc) {
      nxagentParentWindow = (XID) strtol (argv[i], (char**)NULL, 0);
      return 2;
    }
  }

  if (!strcmp(argv[i], "-forcenx")) {
    nxagentForceNXTrans = True;
    return 1;
  }

#ifdef NXAGENT_ONEXIT
  if (!strcmp(argv[i], "-noonce"))
  {
      nxagentOnce = False;
      return 1;
  }
#endif

  if (!strcmp(argv[i], "-kbtype") ||
          !strcmp(argv[i], "-keyboard"))
  {
    if (++i < argc)
    {
      int size;

      if (nxagentKeyboard != NULL)
      {
        xfree(nxagentKeyboard);

        nxagentKeyboard = NULL;
      }

      if ((size = strlen(argv[i])) < 256)
      {
        if ((nxagentKeyboard = xalloc(size + 1)) == NULL)
        {
          FatalError("malloc failed");
        }

        strncpy(nxagentKeyboard, argv[i], size);

        nxagentKeyboard[size] = '\0';
      }
      #ifdef WARNING
      else
      {
        /*
         * it is useless to remember a kbtype
         * option that has just been truncated.
         */

        fprintf(stderr, "ddxProcessArgument: WARNING! Option [%s] too long. "
                    "It will be ignored.\n", argv[i]);
      }
      #endif

      return 2;
    }

    return 0;
  }

  if (!strcmp(argv[i], "-extensions"))
  {
     return 1;
  }

#ifdef RENDER
  if (!strcmp(argv[i], "-norender"))
  {
    nxagentRenderEnable = False;

    return 1;
  }
#endif

  if (!strcmp(argv[i], "-nocomposite"))
  {
    nxagentChangeOption(Composite, 0);

    return 1;
  }

  /*
   * The original -noreset option, disabling
   * dispatchExceptionAtReset, is the default.
   * Use this option to restore the original
   * behaviour.
   */

  if (!strcmp(argv[i], "-reset"))
  {
    nxagentChangeOption(Reset, True);

    return 1;
  }

  if (!strcmp(argv[i], "-persistent"))
  {
    nxagentChangeOption(Persistent, True);

    return 1;
  }

  if (!strcmp(argv[i], "-nopersistent"))
  {
    nxagentChangeOption(Persistent, False);

    return 1;
  }

  if (!strcmp(argv[i], "-noshmem"))
  {
    nxagentChangeOption(SharedMemory, False);

    return 1;
  }

  if (!strcmp(argv[i], "-shmem"))
  {
    nxagentChangeOption(SharedMemory, True);

    return 1;
  }

  if (!strcmp(argv[i], "-noignore"))
  {
    nxagentChangeOption(DeviceControl, True);

    return 1;
  }

  if (!strcmp(argv[i], "-nokbreset"))
  {
    nxagentChangeOption(ResetKeyboardAtResume, False);

    return 1;
  }

  /*
   * Enable pseudo-rootless mode.
   */

  if (!strcmp(argv[i], "-R"))
  {
    if (nxagentOption(Binder) == UNDEFINED ||
            nxagentOption(Desktop) == UNDEFINED ||
                nxagentOption(Rootless) == UNDEFINED)
    {
      nxagentChangeOption(Binder, False);
      nxagentChangeOption(Desktop, False);
      nxagentChangeOption(Rootless, True);
    }
    return 1;
  }

  /*
   * Enable "desktop" mode. This is the
   * default.
   */

  if (!strcmp(argv[i], "-D"))
  {
    nxagentChangeOption(Binder, False);
    nxagentChangeOption(Rootless, False);
    nxagentChangeOption(Desktop, True);

    return 1;
  }

  /*
   * Enable the auto-disconnect timeout.
   */

  if (!strcmp(argv[i], "-timeout"))
  {
    int seconds;

    if (++i < argc && sscanf(argv[i], "%i", &seconds) == 1)
    {
      if (seconds >= 0)
      {
        if (seconds > 0 && seconds < 60)
        {
          seconds = 60;
        }

        nxagentChangeOption(Timeout, seconds);

        return 2;
      }
    }

    return 0;
  }

  /*
   * The return value for -query, -broadcast and
   * -indirect must be 0 to let the dix layer pro-
   * cess these options.
   */

  if (!strcmp(argv[i], "-query"))

  {
    nxagentChangeOption(Desktop, True);
    nxagentChangeOption(Xdmcp, True);
    nxagentMaxAllowedReset = 0;

    return 0;
  }

  if (!strcmp(argv[i], "-broadcast"))

  {
    nxagentChangeOption(Desktop, True);
    nxagentChangeOption(Xdmcp, True);
    nxagentMaxAllowedReset = 0;

    return 0;
  }

  if (!strcmp(argv[i], "-indirect"))
  {
    nxagentChangeOption(Desktop, True);
    nxagentChangeOption(Xdmcp, True);
    nxagentMaxAllowedReset = 1;

    return 0;
  }

  if (!strcmp(argv[i], "-noshpix"))
  {
    nxagentChangeOption(SharedPixmaps, False);

    return 1;
  }

  if (!strcmp(argv[i], "-shpix"))
  {
    nxagentChangeOption(SharedPixmaps, True);

    return 1;
  }

  if (!strcmp(argv[i], "-clipboard"))
  {
    if (!strcmp(argv[i+1], "both"))
    {
      nxagentChangeOption(Clipboard, ClipboardBoth);
    }
    else if (!strcmp(argv[i+1], "client"))
    {
      nxagentChangeOption(Clipboard, ClipboardClient);
    }
    else if (!strcmp(argv[i+1], "server"))
    {
      nxagentChangeOption(Clipboard, ClipboardServer);
    }
    else if (!strcmp(argv[i+1], "none"))
    {
      nxagentChangeOption(Clipboard, ClipboardNone);
    }
    else
    {
      nxagentChangeOption(Clipboard, ClipboardBoth);
    }

    return 2;
  }

  if (!strcmp(argv[i], "-bs"))
  {
    nxagentChangeOption(BackingStore, BackingStoreNever);
    return 1;
  }

  return 0;
}

static void nxagentParseOptions(char *name, char *value)
{
  int size, argc;

  char *argv[2] = { NULL, NULL };

  #ifdef TEST
  fprintf(stderr, "nxagentParseOptions: Processing option '%s' = '%s'.\n",
              validateString(name), validateString(value));
  #endif

  if (!strcmp(name, "kbtype") ||
          !strcmp(name, "keyboard") ||
              !strcmp(name, "id") ||
                  !strcmp(name, "display") ||
                      !strcmp(name, "clipboard") ||
                          !strcmp(name, "geometry") ||
                              !strcmp(name, "option") ||
                                  !strcmp(name, "options") ||
                                      !strcmp(name, "streaming"))
  {
    argv[1] = value;

    argc = 2;
  }
  else if (!strcmp(name, "R") && !strcmp(value, "1"))
  {
    argc = 1;
  }
  else if (!strcmp(name, "fast") || !strcmp(name, "slow"))
  {
    fprintf(stderr, "Warning: Ignoring deprecated option '%s'.\n", name);

    return;
  }
  else if (!strcmp(name, "render"))
  {
    if (nxagentReconnectTrap == True)
    {
      #ifdef DEBUG
      fprintf(stderr, "nxagentParseOptions: Ignoring render option at reconnection.\n");
      #endif
    }
    else if (nxagentRenderEnable == UNDEFINED)
    {
      if (!strcmp(value, "1"))
      {
        nxagentRenderEnable = True;
      }
      else if (!strcmp(value, "0"))
      {
        nxagentRenderEnable = False;
      }
      else
      {
        #ifdef WARNING
        fprintf(stderr, "WARNING: Ignored wrong value [%s] for render option.\n",
                    validateString(value));
        #endif
      }
    }

    return;
  }
  else if (!strcmp(name, "fullscreen"))
  {
    if (nxagentReconnectTrap == True)
    {
      #ifdef DEBUG
      fprintf(stderr, "nxagentParseOptions: Ignoring fullscreen option at reconnection.\n");
      #endif
    }
    else if (!strcmp(value, "1"))
    {
      nxagentChangeOption(Fullscreen, True);
    }
    else if (!strcmp(value, "0"))
    {
      nxagentChangeOption(Fullscreen, False);
    }
    else
    {
      #ifdef WARNING
      fprintf(stderr, "WARNING: Ignored wrong value [%s] for fullscreen option.\n", validateString(value));
      #endif
    }

    return;
  }
  else if (!strcmp(name, "shpix"))
  {
    if (!strcmp(value, "1"))
    {
      nxagentChangeOption(SharedPixmaps, True);
    }
    else if (!strcmp(value, "0"))
    {
      nxagentChangeOption(SharedPixmaps, False);
    }
    else
    {
      #ifdef WARNING
      fprintf(stderr, "WARNING: Ignored wrong value [%s] for shpix option.\n", validateString(value));
      #endif
    }

    return;
  }
  else if (!strcmp(name, "shmem"))
  {
    if (!strcmp(value, "1"))
    {
      nxagentChangeOption(SharedMemory, True);
    }
    else if (!strcmp(value, "0"))
    {
      nxagentChangeOption(SharedMemory, False);
    }
    else
    {
      #ifdef WARNING
      fprintf(stderr, "WARNING: Ignored wrong value [%s] for shmem option.\n", validateString(value));
      #endif
    }

    return;
  }
  else if (!strcmp(name, "composite"))
  {
    if (!strcmp(value, "1"))
    {
      nxagentChangeOption(Composite, 1);
    }
    else if (!strcmp(value, "0"))
    {
      nxagentChangeOption(Composite, 0);
    }
    else
    {
      #ifdef WARNING
      fprintf(stderr, "WARNING: Ignored wrong value [%s] for composite option.\n", validateString(value));
      #endif
    }

    return;
  }
  else if (!strcmp(name, "resize"))
  {
    if (!strcmp(value, "1"))
    {
      nxagentResizeDesktopAtStartup = True;
    }
    else if (!strcmp(value, "0"))
    {
      nxagentResizeDesktopAtStartup = False;
    }
    else
    {
      #ifdef WARNING
      fprintf(stderr, "WARNING: Ignored wrong value [%s] for resize option.\n", validateString(value));
      #endif
    }

    return;
  }
  else if (!strcmp(name, "backingstore"))
  {
    if (!strcmp(value, "0"))
    {
        nxagentChangeOption(BackingStore, BackingStoreNever);
    }
    else
    {
        nxagentChangeOption(BackingStore, BackingStoreForce);
    }

    return;
  }
  else
  {
    #ifdef DEBUG
    fprintf(stderr, "nxagentParseOptionString: Warning ignored option '%s' = '%s'.\n",
                validateString(name), validateString(value));
    #endif

    return;
  }

/*
FIXME: What's the purpose of this? By the way, the memory
       allocated here is not freed.
*/
  if ((size = strlen(name) + 1) > 1)
  {
    if ((argv[0] = malloc(size + 1)) == NULL)
    {
      FatalError("Malloc failed");
    }

    *argv[0] = '-';

    memcpy(argv[0] + 1, name, size);
  }

  ddxProcessArgument(argc, argv, 0);
}

static void nxagentParseOptionString(char *string)
{
  char *value     = NULL;
  char *option    = NULL;
  char *delimiter = NULL;

  /*
   * Remove the port specification.
   */

  delimiter = rindex(string, ':');

  if (delimiter)
  {
    *delimiter = 0;
  }
  else
  {
    #ifdef WARNING
    fprintf(stderr, "nxagentParseOptionString: "
                "Warning option file doesn't contain port specification.\n");
    #endif
  }

  while ((option = strtok(option ? NULL : string, ",")))
  {
    delimiter = rindex(option, '=');

    if (delimiter)
    {
      *delimiter = 0;
      value = delimiter + 1;
    }
    else
    {
      value = NULL;
    }

    nxagentParseOptions(option, value);
  }
}

void nxagentProcessOptionsFile()
{
  FILE *file;
  char *data;

  int offset;
  int size;

  int sizeOfFile;
  int maxFileSize = 1024;

  #ifdef DEBUG
  fprintf(stderr, "nxagentProcessOptionsFile: Going to process option the file [%s].\n",
              validateString(nxagentOptionFile));
  #endif

  if (nxagentOptionFile == NULL)
  {
    return;
  }

  if ((file = fopen(nxagentOptionFile, "r")) == NULL)
  {
    fprintf(stderr, "Warning: Couldn't open option file '%s'. Error is '%s'.\n",
                validateString(nxagentOptionFile), strerror(errno));

    goto nxagentProcessOptionsFileExit;
  }

  if (fseek(file, 0, SEEK_END) != 0)
  {
    fprintf(stderr, "Warning: Couldn't position inside option file '%s'. Error is '%s'.\n",
                validateString(nxagentOptionFile), strerror(errno));

    goto nxagentProcessOptionsFileClose;
  }

  if ((sizeOfFile = ftell(file)) == -1)
  {
    fprintf(stderr, "Warning: Couldn't get the size of option file '%s'. Error is '%s'.\n",
                validateString(nxagentOptionFile), strerror(errno));

    goto nxagentProcessOptionsFileClose;
  }

  #ifdef DEBUG
  fprintf(stderr, "nxagentProcessOptionsFile: Processing file [%s] with size [%d].\n",
              validateString(nxagentOptionFile), strerror(errno));
  #endif

  rewind(file);

  if (sizeOfFile > maxFileSize)
  {
    fprintf(stderr, "Warning: Maximum file size exceeded for options '%s'.\n",
                validateString(nxagentOptionFile));

    goto nxagentProcessOptionsFileClose;
  }

  if ((data = xalloc(sizeOfFile + 1)) == NULL)
  {
    fprintf(stderr, "Warning: Memory allocation failed processing file '%s'.\n",
                validateString(nxagentOptionFile));

    goto nxagentProcessOptionsFileClose;
  }

  offset = 0;
  size = 0;

  for (;;)
  {
    size_t result = fread(data + offset, 1, sizeOfFile, file);

    if (ferror(file) != 0)
    {
      fprintf(stderr, "Warning: Error reading the option file '%s'.\n",
                validateString(nxagentOptionFile));

      goto nxagentProcessOptionsFileFree;
    }

    size   += result;
    offset += result;

    if (feof(file) != 0 || (size == sizeOfFile))
    {
      break;
    }
  }

  if (size != sizeOfFile)
  {
    fprintf(stderr, "Warning: Premature end of option file '%s' while reading.\n",
              validateString(nxagentOptionFile));

    goto nxagentProcessOptionsFileFree;
  }

  /*
   * Truncate the buffer to first line.
   */

  for (offset = 0; (offset < sizeOfFile) && (data[offset] != '\n'); offset++);

  data[offset] = 0;

  nxagentParseOptionString(data);

nxagentProcessOptionsFileFree:

  if (data != NULL)
  {
    Xfree(data);
  }

nxagentProcessOptionsFileClose:

  if (fclose(file) != 0)
  {
    fprintf(stderr, "Warning: Couldn't close option file '%s'. Error is '%s'.\n",
                validateString(nxagentOptionFile), strerror(errno));
  }

nxagentProcessOptionsFileExit:

  return;
}

/*
 * FIXME: Transport initialization, shouldn't depend upon
 *        argv[], because we call it at every reconnection.
 */

Bool nxagentPostProcessArgs(char* name, Display* dpy, Screen* scr)
{
    Bool useNXTrans = False;

    #ifdef WATCH

    fprintf(stderr, "nxagentPostProcessArgs: Watchpoint 2.\n");

/*
Reply   Total	Cached	Bits In			Bits Out		Bits/Reply	  Ratio
------- -----	------	-------			--------		----------	  -----
N/A
*/

    sleep(30);

    #endif

    if ((nxagentOption(Rootless) == 1) && nxagentOption(Fullscreen) == 1)
    {
      #ifdef TEST
      fprintf(stderr, "WARNING: Ignoring fullscreen option for rootless session.\n");
      #endif

      nxagentChangeOption(Fullscreen, False);
    }

    /*
     * Ensure we have a valid name for children dialogs.
     */

    nxagentGetDialogName();

    /*
     * Ensure we have a valid name for window name.
     */

    if (*nxagentWindowName == '\0')
    {
      strncpy(nxagentWindowName, "NX", 255);

      *(nxagentWindowName + 255) = '\0';
    }

    /*
     * Note that use of NX packed images as well as
     * render extension could be later disabled due
     * to the fact that session is running nested
     * in a nxagent server.
     */

    if (nxagentForceNXTrans)
    {
      useNXTrans = True;
    }
    else if ((strncasecmp(name, "nx/", 3) == 0) ||
                 (strncasecmp(name, "nx:", 3) == 0) ||
                     (strncasecmp(name, "nx,", 3) == 0) ||
                         (strcasecmp(name, "nx") == 0))
    {
      useNXTrans = True;
    }

    if (useNXTrans == True)
    {
      unsigned int linkType;
      unsigned int localMajor, localMinor, localPatch;
      unsigned int remoteMajor, remoteMinor, remotePatch;
      int frameTimeout, pingTimeout, splitMode, splitSize;
      unsigned int packMethod, packQuality;
      int dataLevel, streamLevel, deltaLevel;
      unsigned int loadCache, saveCache, startupCache;

      unsigned char supportedMethods[NXNumberOfPackMethods];
      unsigned int entries = NXNumberOfPackMethods;

      unsigned int enableClient;
      unsigned int enableServer;
      unsigned int clientSegment;
      unsigned int serverSegment;
      unsigned int clientSize;
      unsigned int serverSize;

      NXGetControlParameters(dpy, &linkType, &localMajor, &localMinor,
                                 &localPatch, &remoteMajor, &remoteMinor, &remotePatch,
                                     &frameTimeout, &pingTimeout, &splitMode, &splitSize,
                                         &packMethod, &packQuality, &dataLevel, &streamLevel,
                                             &deltaLevel, &loadCache, &saveCache, &startupCache);

      /*
       * This option should be used to verify if the NX
       * transport is enabled, assuming that a value of
       * 0 is also defined as LINK_TYPE_NONE. This is a
       * side effect, in reality, but can be considered
       * safe enough.
       */

      nxagentChangeOption(LinkType, linkType);

      /*
       * If the link is wide enough, we don't
       * need to spare bandwidth.
       */

      if (nxagentOption(LinkType) > LINK_TYPE_ISDN)
      {
        nxagentChangeOption(DefaultLazyLevel, 0);

        nxagentChangeOption(Streaming, 0);
      }

      nxagentSetDefaultLazyLevel();

      /*
       * Setting the synchronization interval
       * based on link type. 
       */
      switch (nxagentOption(LinkType))
      {
        case LINK_TYPE_MODEM:
        {
          nxagentSynchronizationInterval = 1500;

          break;
        }
        case LINK_TYPE_ISDN:
        {
          nxagentSynchronizationInterval = 1000;

          break;
        }
        case LINK_TYPE_ADSL:
        {
          nxagentSynchronizationInterval = 500;

          break;
        }
        case LINK_TYPE_WAN:
        {
          nxagentSynchronizationInterval = 300;

          break;
        }
        case LINK_TYPE_LAN:
        {
          nxagentSynchronizationInterval = 100;

          break;
        }
        default:
        {
          nxagentSynchronizationInterval = 2000;

          break;
        }
      }

      #ifdef TEST
      fprintf(stderr, "nxagentPostProcessArgs: Got local version [%d.%d.%d] remote version [%d.%d.%d].\n",
                  localMajor, localMinor, localPatch, remoteMajor, remoteMinor, remotePatch);

      fprintf(stderr, "nxagentPostProcessArgs: Got karma delay [%d] karma size [%d] split size [%d].\n",
                  karmaDelay, karmaSize, splitSize);

      fprintf(stderr, "nxagentPostProcessArgs: Got preferred pack method [%d] and quality [%d].\n",
                  packMethod, packQuality);
      #endif

      if (nxagentPackMethod == -1)
      {
        nxagentPackMethod = packMethod;
      }

      if (nxagentPackQuality == -1)
      {
        nxagentPackQuality = packQuality;
      }

      if (nxagentTightThreshold == -1)
      {
        nxagentTightThreshold = splitSize;
      }

      if (remoteMajor < 2)
      {
        #ifdef TEST
        fprintf(stderr, "nxagentPostProcessArgs: WARNING! Using backward compatible alpha encoding.\n");
        #endif

        nxagentAlphaCompat = 1;
      }
      else
      {
        nxagentAlphaCompat = 0;
      }

      if (NXGetUnpackParameters(dpy, &entries, supportedMethods) == 0 ||
              entries != NXNumberOfPackMethods)
      {
        fprintf(stderr, "Warning: Unable to retrieve the supported pack methods.\n");

        nxagentPackMethod   = NO_PACK;
        nxagentPackLossless = NO_PACK;
      }
      else
      {
        if (nxagentPackMethod == PACK_RGB_16M_COLORS ||
                nxagentPackMethod == PACK_RLE_16M_COLORS)
        {
          nxagentPackLossless = nxagentPackMethod;
        }
        else
        {
          nxagentPackLossless = PACK_RGB_16M_COLORS;
        }

        if (supportedMethods[nxagentPackLossless] == 0)
        {
          nxagentPackLossless = NO_PACK;
        }

        #ifdef TEST
        fprintf(stderr, "nxagentPostProcessArgs: Using method '%d' for lossless compression.\n",
                    nxagentPackLossless);
        #endif

        if (supportedMethods[nxagentPackMethod] == 0)
        {
          fprintf(stderr, "Warning: Pack method '%d' not supported by proxy.\n",
                      nxagentPackMethod);

          fprintf(stderr, "Warning: Replacing with lossless pack method '%d'.\n",
                      nxagentPackLossless);

          nxagentPackMethod = nxagentPackLossless;
        }

        /*
         * Check if the remote supports the alpha
         * data in compressed form.
         */

        if (nxagentAlphaCompat == 0)
        {
          if (supportedMethods[PACK_ALPHA] == 1)
          {
            #ifdef TEST
            fprintf(stderr, "nxagentPostProcessArgs: Sending alpha data in compressed form.\n");
            #endif

            nxagentAlphaPacked = 1;
          }
          else
          {
            #ifdef TEST
            fprintf(stderr, "nxagentPostProcessArgs: WARNING! Disabling compression of the alpha data.\n");
            #endif

            nxagentAlphaPacked = 0;
          }
        }
      }

      #ifdef TEST
      fprintf(stderr, "nxagentPostProcessArgs: Assuming pack methods [%d] and [%d] with quality [%d].\n",
                  nxagentPackMethod, nxagentPackLossless, nxagentPackQuality);
      #endif

      if (nxagentSplitThreshold == -1)
      {
        nxagentSplitThreshold = splitSize;
      }

      /*
       * Let the remote proxy use the shared memory
       * extension, if supported by the X server.
       * The client part is not useful and not impl-
       * emented. The size of the segment is chosen
       * by the user. The only purpose of the message
       * is to reserve the XID that will be used by
       * the remote.
       */

      enableClient = 0;
      enableServer = 1;

      NXGetShmemParameters(dpy, &enableClient, &enableServer, &clientSegment,
                               &serverSegment, &clientSize, &serverSize);

      if (enableServer == 1)
      {
        fprintf(stderr, "Info: Using shared memory parameters %d/%d/%d/%dK.\n",
                    nxagentOption(SharedMemory), nxagentOption(SharedPixmaps),
                         enableServer, serverSize / 1024);
      }
      else
      {
        fprintf(stderr, "Info: Using shared memory parameters %d/%d/0/0.\n",
                    nxagentOption(SharedMemory), nxagentOption(SharedPixmaps));
      }

      /*
       * We don't need the NoExpose events. Block
       * them at the proxy side.
       */

      NXSetExposeParameters(nxagentDisplay, 1, 1, 0);
    }
    else
    {
      /*
       * We don't have a proxy on the remote side.
       */

      nxagentPackMethod = NO_PACK;

      nxagentSplitThreshold = 0;
    }

    if (nxagentOption(Fullscreen) == UNDEFINED)
    {
      nxagentChangeOption(Fullscreen, False);
    }

    if (nxagentOption(Binder) == UNDEFINED)
    {
      nxagentChangeOption(Binder, False);
    }

    if (nxagentOption(Rootless) == UNDEFINED)
    {
      nxagentChangeOption(Rootless, False);
    }

    if (nxagentOption(Desktop) == UNDEFINED)
    {
      nxagentChangeOption(Desktop, True);
    }

/*
FIXME: In rootless mode the backing-store support is not functional yet.
*/
    if (nxagentOption(Rootless))
    {
      enableBackingStore = 0;
    }
    else if (nxagentOption(BackingStore) == BackingStoreUndefined ||
                 nxagentOption(BackingStore) == BackingStoreForce)
    {
      enableBackingStore = 1;
    }
    else if (nxagentOption(BackingStore) == BackingStoreNever)
    {
      enableBackingStore = 0;
    }

    /*
     * need to check if this was set on the
     * command line as this has the priority
     * over the option file.
     */
 
    if (nxagentRenderEnable == UNDEFINED)
    {
      nxagentRenderEnable = True;
    }
    
    if (nxagentRenderEnable == True)
    {
      nxagentAlphaEnabled = True;
    }
    else
    {
      nxagentAlphaEnabled = False;
    }

    if ((nxagentOption(Rootless) == 1) && nxagentOption(Xdmcp))
    {
      FatalError("PANIC! Cannot start a XDMCP session in rootless mode.\n");
    }

    /*
     * We enable server reset only for indirect XDMCP sessions.
     */

    if (nxagentOption(Reset) == True && nxagentMaxAllowedReset == 0)
    {
      #ifdef WARNING
      fprintf(stderr, "nxagentPostProcessArgs: Disabling reset.\n");
      #endif

      nxagentChangeOption(Reset, False);

      dispatchExceptionAtReset = 0;
    }

    /*
     * We skip server reset by default. This should
     * be equivalent to passing the -noreset option
     * to a standard XFree86 server.
     */

    if (nxagentOption(Reset) == False)
    {
      #ifdef TEST
      fprintf(stderr, "nxagentPostProcessArgs: Disabling dispatch of exception at server reset.\n");
      #endif

      dispatchExceptionAtReset = 0;
    }

    /*
     * Check if the user activated the auto-disconect
     * feature.
     */

    if (nxagentOption(Timeout) > 0)
    {
      fprintf(stderr, "Info: Using auto-disconnect timeout of %d seconds.\n",
                  nxagentOption(Timeout));

      nxagentAutoDisconnectTimeout = nxagentOption(Timeout) * MILLI_PER_SECOND;
    }

    #ifdef WATCH

    fprintf(stderr, "nxagentPostProcessArgs: Watchpoint 3.\n");

/*
Reply   Total	Cached	Bits In			Bits Out		Bits/Reply	  Ratio
------- -----	------	-------			--------		----------	  -----
#16     1		256 bits (0 KB) ->	12 bits (0 KB) ->	256/1 -> 12/1	= 21.333:1
#233 A  1		256 bits (0 KB) ->	131 bits (0 KB) ->	256/1 -> 131/1	= 1.954:1
#245 A  2		512 bits (0 KB) ->	19 bits (0 KB) ->	256/1 -> 10/1	= 26.947:1
*/

    sleep(30);

    #endif

    return useNXTrans;
}

void ddxUseMsg()
{
  ErrorF("-display string        display name of the real server\n");
  ErrorF("-sync                  synchronize with the real server\n");
  ErrorF("-full                  utilize full regeneration\n");
  ErrorF("-class string          default visual class\n");
  ErrorF("-depth int             default depth\n");
  ErrorF("-geometry WxH+X+Y      window size and position\n");
  ErrorF("-bw int                window border width\n");
  ErrorF("-name string           window name\n");
  ErrorF("-scrns int             number of screens to generate\n");
  ErrorF("-install               install colormaps directly\n");

  ErrorF("The NX system adds the following arguments:\n");
  ErrorF("-forcenx               force use of NX protocol messages assuming communication through nxproxy\n");
  ErrorF("-timeout int           auto-disconnect timeout in seconds (minimum allowed: 60)\n");
#ifdef NXAGENT_XLIBPARAMS
  ErrorF("-imgframe              let NXlib draw a tiny rectangle around images\n");
  ErrorF("-imgsplit PW,PH,BW,BH  force NXlib to split images to the given width and height\n");
#endif
#ifdef RENDER
  ErrorF("-norender              disable the use of the render extension\n");
  ErrorF("-nocomposite           disable the use of the composite extension\n");
#endif
  ErrorF("-nopersistent          disable disconnection/reconnection to the X display on SIGHUP\n");
  ErrorF("-noshmem               disable use of shared memory extension\n");
  ErrorF("-shmem                 enable use of shared memory extension\n");
  ErrorF("-noshpix               disable use of shared pixmaps\n");
  ErrorF("-shpix                 enable use of shared pixmaps\n");
  ErrorF("-noignore              don't ignore pointer and keyboard configuration changes mandated by clients\n");
  ErrorF("-nokbreset             don't reset keyboard device if the session is resumed\n");
  ErrorF("-D                     enable desktop mode\n");
  ErrorF("-R                     enable rootless mode\n");
  ErrorF("-B                     enable proxy binding mode\n");
}

static int nxagentGetDialogName()
{
  strcpy(nxagentDialogName, "NX");

  *(nxagentDialogName + 255) = '\0';

  if (*nxagentSessionId != '\0')
  {
    int length = strlen(nxagentSessionId);

    strcpy(nxagentDialogName, "NX - ");

    if (length > (MD5_LENGTH * 2 + 1) &&
           *(nxagentSessionId + (length - (MD5_LENGTH * 2 + 1))) == '-')
    {
      strncat(nxagentDialogName, nxagentSessionId, length - (MD5_LENGTH * 2 + 1));
    }
    else
    {
      strncat(nxagentDialogName, nxagentSessionId, 250);
    }

    *(nxagentSessionId + 255) = '\0';

    return 1;
  }

  return 0;
}

/*
 * Each lazy level adds the following rules to the previous ones:
 *
 * Level 0  The lazy is off.
 *
 * Level 1  The put images are skipped if we were out of bandwidth,
 *          unless that the destination drawable has an old corru-
 *          pted region.
 *
 * Level 2  No data is put or copied on pixmaps, marking them always
 *          as corrupted and synchronizing them on demand.
 *
 * Level 3  The put images over the windows are skipped marking the
 *          destination as corrupted. When a copy area to a window is
 *          requested, the source is synchronized before copying it.
 *
 * Level 4  The source drawable is no longer synchronized before a
 *          copy area, but the operation is clipped to the synchro-
 *          nized region.
 */

void nxagentSetDefaultLazyLevel()
{
  switch (nxagentOption(LinkType))
  {
    case LINK_TYPE_MODEM:
    case LINK_TYPE_ISDN:
    {
      nxagentChangeOption(DefaultLazyLevel, 2);

      break;
    }

    case LINK_TYPE_ADSL:
    case LINK_TYPE_WAN:
    case LINK_TYPE_LAN:
    default:
    {
      nxagentChangeOption(DefaultLazyLevel, 0);

      break;
    }
  }

  #ifdef TEST
  fprintf(stderr, "nxagentSetDefaultLazyLevel: Default lazy level set to [%d].\n",
              nxagentOption(DefaultLazyLevel));
  #endif
}

