/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2001, 2006 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* NXAGENT, NX protocol compression and NX extensions to this software    */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for applicability.       */
/*                                                                        */
/* NX and NoMachine are trademarks of Medialogic S.p.A.                   */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

/*

Copyright 1993 by Davor Matic

Permission to use, copy, modify, distribute, and sell this software
and its documentation for any purpose is hereby granted without fee,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation.  Davor Matic makes no representations about
the suitability of this software for any purpose.  It is provided "as
is" without express or implied warranty.

*/

#include <signal.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <unistd.h>
#include <time.h>
#include <errno.h>

#include <X11/X.h>
#include <X11/Xproto.h>
#include "screenint.h"
#include "input.h"
#include "misc.h"
#include "scrnintstr.h"
#include "servermd.h"
#include "windowstr.h"

#ifdef WATCH
#include "unistd.h"
#endif

#include "NXalert.h"

#include "Agent.h"
#include "Display.h"
#include "Visual.h"
#include "Options.h"
#include "Error.h"
#include "Init.h"
#include "Args.h"
#include "Icons.h"
#include "Render.h"
#include "Font.h"
#include "Reconnect.h"
#include "Events.h"
#include "Dialog.h"
#include "Client.h"
#include "Splash.h"
#include "Screen.h"

#include "NX.h"
#include "NXlib.h"

#include NXAGENT_ICON_NAME

/*
 * Set here the required log level.
 */

#define PANIC
#define WARNING
#undef  TEST
#undef  DEBUG
#undef  WATCH

Display *nxagentDisplay = NULL;
XVisualInfo *nxagentVisuals = NULL;
Bool nxagentTrue24 = False;

int nxagentNumVisuals;
int nxagentXConnectionNumber;

/*
 * Set if we had to block waiting for
 * the display to become writable.
 */

int nxagentBlocked = 0;

int nxagentIOErrorHandler(Display *display);

static Bool nxagentDisplayInfoSaved = False;
static Display *nxagentDisplayBackup = NULL;
static XlibGC nxagentBitmapGCBackup = NULL;
static XVisualInfo *nxagentVisualsRecBackup;
static int nxagentNumVisualsRecBackup;
static int nxagentNumDefaultColormapsRecBackup;
static int *nxagentDepthsRecBackup;
static int nxagentNumDepthsRecBackup;
static int nxagentNumPixmapFormatsRecBackup;
static XPixmapFormatValues *nxagentPixmapFormatsRecBackup;
static int DisplayWidthRecBackup;
static int DisplayHeightRecBackup;
static Bool nxagentRenderEnableRecBackup;
static Bool *nxagentVisualHasBeenIgnored;

static enum
{
  NOTHING = 0,
  OPENED,
  GOT_VISUAL_INFO,
  ALLOC_DEF_COLORMAP,
  GOT_DEPTH_LIST,
  GOT_PIXMAP_FORMAT_LIST,
  EVERYTHING_DONE
} reconnectDisplayState;

int nxagentDefaultVisualIndex;
Colormap *nxagentDefaultColormaps;
int nxagentNumDefaultColormaps;
int *nxagentDepths;
int nxagentNumDepths;
XPixmapFormatValues *nxagentPixmapFormats;
int nxagentNumPixmapFormats;
Pixel nxagentBlackPixel;
Pixel nxagentWhitePixel;
Drawable nxagentDefaultDrawables[MAXDEPTH + 1];
Pixmap nxagentScreenSaverPixmap;

/*
 * Also used in Cursor.c. There are huge problems
 * using GC definition. This is to be reworked.
 */

XlibGC nxagentBitmapGC;

/*
 * The "confine" window is used in the nxagentConstrainCursor
 * procedure. We are currently overriding the original Xnest
 * behaviour. It is unclear what this window is used for.
 */

Window nxagentConfineWindow;

Pixmap nxagentIconPixmap;
Pixmap nxagentIconShape;
Bool useXpmIcon = False;

unsigned int nxagentLogoColor(unsigned int value);
Bool nxagentMakeIcon(Display *display, Pixmap *nxIcon, Pixmap *nxMask);

/*
 * FIXME: These must definitely become local.
 */

XVisualInfo pV;
unsigned int r, b, g, or, ob, og, off;

/*
 * Other local functions.
 */

Display *nxagentInternalOpenDisplay(char *display);
void nxagentRaiseSignal(int signal);
void nxagentDisconnectDisplay(void);
Bool nxagentReconnectDisplay(void *p0);

#ifdef NXAGENT_TIMESTAMP
unsigned long startTime;
#endif

/*
 * This is located in connection.c.
 */

extern void RejectWellKnownSockets(void);

int nxagentServerOrder()
{
  int whichbyte = 1;

  if (*((char *) &whichbyte))
      return LSBFirst;

  return MSBFirst;
}

unsigned int nxagentLogoColor(unsigned int value)
{
  /*
   * Takes a color value in RGB24 (0xff0000, 0x00ff00,
   * 0x0000ff) and converts it into an equivalent for
   * the current visual.
   */

  int cr=0, cg=0, cb=0;

  cr = (value >> or) &r;
  cg = (value >> (og - 8)) &g;
  cb = (value >> (ob - 16)) &b;

  return (cr | cg | cb);
}

/*
 * FIXME: This error handler is not printing anything
 * in the session log. This is OK once the session is
 * started, because the error is handled by the other
 * layers, but not before that point, as the agent
 * would die without giving any feedback to the user
 * (or, worse, to the NX server). We should check how
 * many requests have been handled for this display
 * and print a message if the display dies before the
 * session is up and running.
 */

/*
 * FIXME: This should be moved to Error.c, The other
 * handlers should be probably moved to Handlers.c.
 */

int nxagentIOErrorHandler(Display *display)
{
  #ifdef TEST
  fprintf(stderr, "nxagentIOErrorHandler: Got I/O error with nxagentException.ioError [%d].\n",
              nxagentException.ioError);
  #endif

  nxagentException.ioError++;
 
  #ifdef TEST
  fprintf(stderr, "nxagentIOErrorHandler: Set nxagentException.ioError to [%d].\n",
              nxagentException.ioError);
  #endif

  return 1;
}

/*
 * Force a shutdown of any connection attempt
 * while connecting to the remote display.
 * This is needed to avoid a hang up in case
 * of loopback connections to our own listen-
 * ing sockets.
 */

static void nxagentRejectConnection(int signal)
{
  #ifdef TEST
  fprintf(stderr, "nxagentRejectConnection: Going to reject client connections.\n");
  #endif

  #ifndef NXAGENT_UPGRADE

  /*
   * FIXME: The port to the Xorg tree is still a work
   * in progress. This function is not available in
   * the new tree.
   */

  RejectWellKnownSockets();

  #endif

  #ifdef TEST
  fprintf(stderr, "nxagentRejectConnection: Setting new alarm to 5 seconds from now.\n");
  #endif

  /*
   * A further timeout is unlikely to happen
   * in the case of loopback connections.
   */

  alarm(5);
}

/*
 * Ignore the signal if the NX transport is
 * not running.
 */

static void nxagentSigusrHandler(int signal)
{
  #ifdef TEST
  fprintf(stderr, "nxagentSigusrHandler: Nothing to do with signal [%d].\n",
              signal);
  #endif
}

static void nxagentSighupHandler(int signal)
{
  #ifdef TEST
  fprintf(stderr, "nxagentSighupHandler: Handling signal with state [%s] transport [%d] server "
              "generation [%ld].\n", DECODE_SESSION_STATE(nxagentSessionState),
                  NXTransRunning(NX_FD_ANY), serverGeneration);
  #endif

  if (signal != SIGHUP)
  {
    #ifdef PANIC
    fprintf(stderr, "nxagentSighupHandler: PANIC! Invalid signal [%d] received in state [%s].\n",
                signal, DECODE_SESSION_STATE(nxagentSessionState));
    #endif

    return;
  }

  if (dispatchException & DE_TERMINATE)
  {
    #ifdef TEST
    fprintf(stderr, "nxagentSighupHandler: Ignoring the signal while terminating the session.\n");
    #endif

    return;
  }
  else if (nxagentSessionState == SESSION_UP)
  {
    if (nxagentOption(Persistent) == 1)
    {
      #ifdef TEST
      fprintf(stderr, "nxagentSighupHandler: Handling the signal by disconnecting the agent.\n");
      #endif

      nxagentException.sigHup++;
    }
    else
    {
      #ifdef TEST
      fprintf(stderr, "nxagentSighupHandler: Ignoring the signal with persistency disabled.\n");
      #endif
    }

    return;
  }
  else if (nxagentSessionState == SESSION_STARTING)
  {
    #ifdef TEST
    fprintf(stderr, "nxagentSighupHandler: Handling the signal by aborting the session.\n");
    #endif

    nxagentException.sigHup++;

    return;
  }
  else if (nxagentSessionState == SESSION_DOWN)
  {
    if (NXTransRunning(NX_FD_ANY) == 1)
    {
      #ifdef TEST
      fprintf(stderr, "nxagentSighupHandler: Handling the signal by aborting the reconnection.\n");
      #endif
    }
    else
    {
      #ifdef TEST
      fprintf(stderr, "nxagentSighupHandler: Handling the signal by resuming the session.\n");
      #endif
    }

    nxagentException.sigHup++;

    return;
  }

  #ifdef WARNING
  fprintf(stderr, "nxagentSighupHandler: WARNING! Ignoring the signal in state [%s].\n",
              DECODE_SESSION_STATE(nxagentSessionState));
  #endif
}

static void nxagentSigchldHandler(int signal)
{
  int pid = 0;
  int status;
  int options;

  #ifdef TEST
  fprintf(stderr, "nxagentSigchldHandler: Going to check the children processes.\n");
  #endif

  options = WNOHANG | WUNTRACED;

  /*
   * Try with the pid of the dialog process.
   * Leave the other children unaffected.
   */

  if (pid == 0 && nxagentRootlessDialogPid)
  {
    pid = waitpid(nxagentRootlessDialogPid, &status, options);

    if (pid == -1 && errno == ECHILD)
    {
      #ifdef WARNING
      fprintf(stderr, "nxagentSigchldHandler: Got ECHILD waiting for child %d (Rootless dialog).\n",
                  nxagentRootlessDialogPid);
      #endif

      pid = nxagentRootlessDialogPid = 0;
    }
  }

  if (pid == 0 && nxagentPulldownDialogPid)
  {
    pid = waitpid(nxagentPulldownDialogPid, &status, options);

    if (pid == -1 && errno == ECHILD)
    {
      #ifdef WARNING
      fprintf(stderr, "nxagentSigchldHandler: Got ECHILD waiting for child %d (Pulldown dialog).\n",
                  nxagentPulldownDialogPid);
      #endif

      pid = nxagentPulldownDialogPid = 0;
    }
  }

  if (pid == 0 && nxagentKillDialogPid)
  {
    pid = waitpid(nxagentKillDialogPid, &status, options);

    if (pid == -1 && errno == ECHILD)
    {
      #ifdef WARNING
      fprintf(stderr, "nxagentSigchldHandler: Got ECHILD waiting for child %d (Kill dialog).\n",
                  nxagentKillDialogPid);
      #endif

      pid = nxagentKillDialogPid = 0;
    }
  }

  if (pid == 0 && nxagentSuspendDialogPid)
  {
    pid = waitpid(nxagentSuspendDialogPid, &status, options);

    if (pid == -1 && errno == ECHILD)
    {
      #ifdef WARNING
      fprintf(stderr, "nxagentSigchldHandler: Got ECHILD waiting for child %d (Suspend dialog).\n",
                  nxagentSuspendDialogPid);
      #endif

      pid = nxagentSuspendDialogPid = 0;
    }
  }

  if (pid == 0 && nxagentEnableRandRModeDialogPid)
  {
    pid = waitpid(nxagentEnableRandRModeDialogPid, &status, options);

    if (pid == -1 && errno == ECHILD)
    {
      #ifdef WARNING
      fprintf(stderr, "nxagentSigchldHandler: Got ECHILD waiting for child %d (EnableRandRMode dialog).\n",
                  nxagentEnableRandRModeDialogPid);
      #endif

      pid = nxagentEnableRandRModeDialogPid = 0;
    }
  }

  if (pid == 0 && nxagentDisableRandRModeDialogPid)
  {
    pid = waitpid(nxagentDisableRandRModeDialogPid, &status, options);

    if (pid == -1 && errno == ECHILD)
    {
      #ifdef WARNING
      fprintf(stderr, "nxagentSigchldHandler: Got ECHILD waiting for child %d (DisableRandRMode dialog).\n",
                  nxagentDisableRandRModeDialogPid);
      #endif

      pid = nxagentDisableRandRModeDialogPid = 0;
    }
  }

  if (pid == 0 && nxagentEnableLazyModePid)
  {
    pid = waitpid(nxagentEnableLazyModePid, &status, options);

    if (pid == -1 && errno == ECHILD)
    {
      #ifdef WARNING
      fprintf(stderr, "nxagentSigchldHandler: Got ECHILD waiting for child %d (EnableLazyMode dialog).\n",
                  nxagentEnableLazyModePid);
      #endif

      pid = nxagentEnableLazyModePid = 0;
    }
  }

  if (pid == 0 && nxagentDisableLazyModePid)
  {
    pid = waitpid(nxagentDisableLazyModePid, &status, options);

    if (pid == -1 && errno == ECHILD)
    {
      #ifdef WARNING
      fprintf(stderr, "nxagentSigchldHandler: Got ECHILD waiting for child %d (DisableLazyMode dialog).\n",
                  nxagentDisableLazyModePid);
      #endif

      pid = nxagentDisableLazyModePid = 0;
    }
  }

  if (pid == -1)
  {
    FatalError("nxagentSigchldHandler: Got error %s waiting for child.\n", strerror(errno));
  }

  if (pid > 0)
  {
    if (WIFSTOPPED(status))
    {
      #ifdef WARNING
      fprintf(stderr, "nxagentSigchldHandler: Child process [%d] was stopped "
                  "with signal [%d].\n", pid, (WSTOPSIG(status)));
      #endif
    }
    else
    {
      #ifdef TEST

      if (WIFEXITED(status))
      {
        fprintf(stderr, "nxagentSigchldHandler: Child process [%d] exited "
                    "with status [%d].\n", pid, (WEXITSTATUS(status)));
      }
      else if (WIFSIGNALED(status))
      {
        fprintf(stderr, "nxagentSigchldHandler: Child process [%d] died "
                    "because of signal [%d].\n", pid, (WTERMSIG(status)));
      }

      #endif

      nxagentResetDialog(pid);
    }
  }
  #ifdef TEST
  else if (pid == 0)
  {
    fprintf(stderr, "nxagentSigchldHandler: No own child process exited.\n");
  }
  #endif

  return;
}

Display *nxagentInternalOpenDisplay(char *display)
{
  Display *newDisplay;

  struct sigaction oldAction;
  struct sigaction newAction;

  int result;

  /*
   * By default nxcomp installs its signal handlers.
   * We need to ensure that SIGUSR1 and SIGUSR2 are
   * ignored if the NX transport is not running.
   */

  newAction.sa_handler = nxagentSigusrHandler;

  sigfillset(&newAction.sa_mask);

  newAction.sa_flags = SA_RESTART;

  while (((result = sigaction(SIGUSR1, &newAction, NULL)) == -1) &&
             (errno == EINTR));

  if (result == -1)
  {
    FatalError("Error: Can't set handler for user signal 1.");
  }

  while (((result = sigaction(SIGUSR2, &newAction, NULL)) == -1) &&
             (errno == EINTR));

  if (result == -1)
  {
    FatalError("Error: Can't set handler for user signal 2.");
  }

  /*
   * Install the handler rejecting a possible
   * loopback connection.
   */

  newAction.sa_handler = nxagentRejectConnection;

  sigfillset(&newAction.sa_mask);

  newAction.sa_flags = SA_RESTART;

  while (((result = sigaction(SIGALRM, &newAction, &oldAction)) == -1) &&
             (errno == EINTR));

  if (result == -1)
  {
    FatalError("Error: Can't set alarm for rejecting connections.");
  }

  alarm(10);

  newDisplay = XOpenDisplay(display);

  alarm(0);

  while (((result = sigaction(SIGALRM, &oldAction, NULL)) == -1) &&
             (errno == EINTR));

  if (result == -1)
  {
    FatalError("Error: Can't restore alarm for rejecting connections.");
  }

  /*
   * Let the default X server's handler manage the
   * SIGINT and SIGTERM. The following calls will
   * either restore the old handlers (those origin-
   * ally installed by the X server) or will not do
   * anything, if the NX transport is not running.
   */

  NXTransSignal(SIGINT,  NX_SIGNAL_DISABLE);
  NXTransSignal(SIGTERM, NX_SIGNAL_DISABLE);

  /*
   * Also tell the proxy to ignore the SIGHUP signal
   * and install our own handler.
   */

  NXTransSignal(SIGHUP, NX_SIGNAL_DISABLE);

  newAction.sa_handler = nxagentSighupHandler;

  sigfillset(&newAction.sa_mask);

  newAction.sa_flags = SA_RESTART;

  while (((result = sigaction(SIGHUP, &newAction, NULL)) == -1) &&
             (errno == EINTR));

  if (result == -1)
  {
    FatalError("Error: Can't set the signal handler for reconnection.");
  }

  /*
   * Both the proxy and the agent need to be notified
   * about their children, so let the proxy install
   * its handler and forward the signal to us.
   */

  newAction.sa_handler = nxagentSigchldHandler;

  sigfillset(&newAction.sa_mask);

  newAction.sa_flags = SA_RESTART;

  while (((result = sigaction(SIGCHLD, &newAction, NULL)) == -1) &&
             (errno == EINTR));

  if (result == -1)
  {
    FatalError("Error: Can't set signal handler for children.");
  }

  NXTransSignal(SIGCHLD, NX_SIGNAL_FORWARD);

  /*
   * For now let the proxy flush the link as soon as
   * there is encoded data. We will change the policy
   * before entering the dispatcher.
   */
/*
FIXME: It may be a good idea to set the flush policy
       to deferred just after having opened the display
       and flush the link explicitly in the display
       block handler.
*/

  #ifdef TEST
  fprintf(stderr, "nxagentInternalOpenDisplay: Setting the NX flush policy to immediate.\n");
  #endif

  NXSetDisplayPolicy(nxagentDisplay, NXPolicyImmediate);

  #ifdef TEST
  fprintf(stderr, "nxagentInternalOpenDisplay: Successfully opened the display at [%p].\n",
              (void *) newDisplay);
  #endif

  return newDisplay;
}

void nxagentRaiseSignal(int signal)
{
  if (kill(getpid(), signal) == -1)
  {
    #ifdef WARNING
    fprintf(stderr, "nxagentRaiseSignal: Couldn't raise signal %d: %s.\n",
                signal, strerror(errno));
    #endif
  }
}

static void nxagentDisplayFlushHandler(Display *display, int reason)
{
  #ifdef TEST
  fprintf(stderr, "nxagentDisplayFlushHandler: FLUSH! Flush requested on the proxy link.\n");
  #endif

  nxagentFlush = 1;
}

static void nxagentDisplayCongestionHandler(Display *display, int reason)
{
  if (reason == NXBeginCongestion)
  {
    #ifdef TEST
    fprintf(stderr, "nxagentDisplayCongestionHandler: CONGESTION! Congestion state is [begin].\n");
    #endif

    nxagentCongestion = 1;
  }
  else
  {
    #ifdef TEST
    fprintf(stderr, "nxagentDisplayCongestionHandler: CONGESTION! Congestion state is [end].\n");
    #endif

    nxagentCongestion = 0;
  }
}

static void nxagentDisplaySynchronizationHandler(Display *display, int reason)
{
  if (reason == NXBeginSynchronization)
  {
    #ifdef TEST
    fprintf(stderr, "nxagentDisplaySynchronizationHandler: SYNCHRONIZATION! Synchronization state is [begin].\n");
    #endif

    nxagentSynchronization = 1;
  }
  else
  {
    #ifdef TEST
    fprintf(stderr, "nxagentDisplaySynchronizationHandler: SYNCHRONIZATION! Synchronization state is [end].\n");
    #endif

    nxagentSynchronization = 0;
  }
}

static void nxagentDisplayBlockHandler(Display *display, int reason)
{
  if (nxagentDisplay != NULL)
  {
    #ifdef TEST

    int pending = NXQueryDisplay(nxagentDisplay, NXDisplayFlushable);

    if (pending > 0)
    {
      fprintf(stderr, "nxagentDisplayBlockHandler: WARNING! BLOCK! The NX transport has [%d] "
                  "bytes to flush with reason [%d].\n", pending, reason);
    }

    #endif

    if (reason == NXBlockRead)
    {
      #ifdef DEBUG
      fprintf(stderr, "nxagentDisplayBlockHandler: BLOCK! Display is blocking for [read].\n");
      #endif

      NXFlushDisplay(nxagentDisplay, NXFlushIdle);
    }
    else
    {
      #ifdef DEBUG
      fprintf(stderr, "nxagentDisplayBlockHandler: BLOCK! Display is blocking for [write].\n");
      #endif

      NXFlushDisplay(nxagentDisplay, NXFlushLink);

      nxagentBlocked = 1;
    }
  }
}

static int nxagentDisplayErrorPredicate(Display *display, int error)
{
/*
FIXME: This is called too many times in the same loop.
       We should find a way to save the function call.
*/
  #ifdef TEST
  fprintf(stderr, "nxagentDisplayErrorPredicate: CHECK! Error is [%d] with [%d][%d][%d][%d][%d].\n",
              ((error == 1) || (dispatchException & DE_RESET) != 0 ||
                  (dispatchException & DE_TERMINATE) != 0 || nxagentException.sigHup > 0 ||
                      nxagentException.ioError > 0), error, (dispatchException & DE_RESET) != 0,
                          (dispatchException & DE_TERMINATE), nxagentException.sigHup > 0,
                              nxagentException.ioError > 0);
  #endif

  if (error == 0)
  {
    if ((dispatchException & DE_RESET) != 0 ||
            (dispatchException & DE_TERMINATE) != 0 ||
                nxagentException.sigHup > 0 ||
                    nxagentException.ioError > 0)
    {
      return NXForceDisplayError(display);
    }
  }

  return error;
}

void nxagentInstallDisplayHandlers()
{
  /*
   * If the display was already opened, free
   * the internal nxcompext structures.
   */

  if (nxagentDisplay != NULL)
  {
    NXResetDisplay(nxagentDisplay);

    /*
     * Reset the congestion state.
     */

    nxagentCongestion = False;
  }

  /*
   * We want the Xlib I/O error handler to return,
   * instead of quitting the application. Using
   * setjmp()/longjmp() leaves the door open to
   * unexpected bugs when dealing with interaction
   * with the other X server layers.
   */

  NXHandleDisplayError(1);

  NXSetDisplayFlushHandler(nxagentDisplayFlushHandler, NULL);

  NXSetDisplayCongestionHandler(nxagentDisplayCongestionHandler, NULL);

  NXSetDisplaySynchronizationHandler(nxagentDisplaySynchronizationHandler, NULL);

  NXSetDisplayBlockHandler(nxagentDisplayBlockHandler);

  /*
   * Override the default Xlib error handler.
   */

  XSetIOErrorHandler(nxagentIOErrorHandler);

  /*
   * Let Xlib become aware of our interrupts. In theory
   * we don't need to have the error handler installed
   * during the normal operations and could simply let
   * the dispatcher handle the interrupts. In practice
   * it's better to have Xlib invalidating the display
   * as soon as possible rather than incurring in the
   * risk of entering a loop that doesn't care checking
   * the display errors explicitly.
   */

  #ifdef TEST
  fprintf(stderr, "nxagentInstallDisplayHandlers: Installing the error function predicate.\n");
  #endif

  NXSetDisplayErrorPredicate(nxagentDisplayErrorPredicate);
}

void nxagentPostInstallDisplayHandlers()
{
  /*
   * This is executed after having opened the
   * display, once we know the display address.
   */

  if (nxagentDisplay != NULL)
  {
    #ifdef TEST
    fprintf(stderr, "nxagentPostInstallDisplayHandlers: Initializing the NX display internals.\n");
    #endif

    NXInitDisplay(nxagentDisplay);
/*
FIXME: What is the most appropriate number of elements?

    NXInitCache(nxagentDisplay, 128);
*/
    NXInitCache(nxagentDisplay, 256);

    NXSetDisplayFlushHandler(nxagentDisplayFlushHandler, nxagentDisplay);

    NXSetDisplayCongestionHandler(nxagentDisplayCongestionHandler, nxagentDisplay);

    NXSetDisplaySynchronizationHandler(nxagentDisplaySynchronizationHandler, nxagentDisplay);
  }

  /*
   * Handler for the Xlib protocol errors.
   */

  XSetErrorHandler(nxagentErrorHandler);
}

void nxagentOpenDisplay(int argc, char *argv[])
{
  XVisualInfo vi;
  XVisualInfo *viList = NULL;
  long mask;
  int i, viNumList;

  if (!nxagentDoFullGeneration) return;

#ifdef NXAGENT_TIMESTAMP
  startTime = GetTimeInMillis();
  fprintf(stderr, "Display: Going to open Display on real X server time is [%d] milliseconds.\n",
          GetTimeInMillis() - startTime);
#endif

  /*
   * Initialize the reconnector only in the case
   * of persistent sessions.
   */

  if (nxagentOption(Persistent))
  {
    nxagentInitReconnector();
  }

  if (*nxagentDisplayName == '\0')
  {
    strncpy(nxagentDisplayName, XDisplayName(NULL), 1023);

    nxagentDisplayName[1023] = '\0';
  }

  nxagentCloseDisplay();

  nxagentInstallDisplayHandlers();

  nxagentDisplay = nxagentInternalOpenDisplay(nxagentDisplayName);

  nxagentPostInstallDisplayHandlers();

  if (nxagentDisplay == NULL)
  {
/*
FIXME: The agent should never exit the program with a FatalError()
       but rather use a specific function that may eventually call 
       FatalError() on its turn.
*/
    FatalError("Error: Unable to open display '%s'.\n", nxagentDisplayName);
  }

  nxagentXConnectionNumber = XConnectionNumber(nxagentDisplay);

  #ifdef TEST
  fprintf(stderr, "nxagentOpenDisplay: Display image order is [%d] bitmap order is [%d].\n",
              ImageByteOrder(nxagentDisplay), BitmapBitOrder(nxagentDisplay));

  fprintf(stderr, "nxagentOpenDisplay: Display scanline unit is [%d] scanline pad is [%d].\n",
              BitmapUnit(nxagentDisplay), BitmapPad(nxagentDisplay));
  #endif

  #ifdef WATCH

  fprintf(stderr, "nxagentOpenDisplay: Watchpoint 1.\n");

/*
Reply   Total	Cached	Bits In			Bits Out		Bits/Reply	  Ratio
------- -----	------	-------			--------		----------	  -----
#1   U  1	1	256 bits (0 KB) ->	150 bits (0 KB) ->	256/1 -> 150/1	= 1.707:1
#20     1	1	119104 bits (15 KB) ->	28 bits (0 KB) ->	119104/1 -> 28/1	= 4253.714:1
#98     2		512 bits (0 KB) ->	84 bits (0 KB) ->	256/1 -> 42/1	= 6.095:1
*/

  sleep(60);

  #endif

#ifdef NXAGENT_TIMESTAMP
  fprintf(stderr, "Display: display on real X opened, time is [%d] milliseconds.\n",
          GetTimeInMillis() - startTime);
#endif

  nxagentUseNXTrans =
      nxagentPostProcessArgs(nxagentDisplayName, nxagentDisplay,
                                 DefaultScreenOfDisplay(nxagentDisplay));

  /*
   * Processing the arguments all the timeouts
   * have been set. Now we have to change the
   * screen-saver timeout.
   */

  nxagentSetScreenSaverTime();

  mask = VisualScreenMask;
  vi.screen = DefaultScreen(nxagentDisplay);
  vi.depth = DefaultDepth(nxagentDisplay, DefaultScreen(nxagentDisplay));
  viList = XGetVisualInfo(nxagentDisplay, mask, &vi, &viNumList);
  nxagentVisuals = (XVisualInfo *) malloc(viNumList * sizeof(XVisualInfo));
  nxagentNumVisuals = 0;

  for (i = 0; i < viNumList; i++)
  {
    if (viList[i].depth == vi.depth)
    {
      if (nxagentVisuals != NULL)
      {
        memcpy(nxagentVisuals + nxagentNumVisuals, viList + i, sizeof(XVisualInfo));
      }

      nxagentNumVisuals++;
    }
  }

  if (nxagentVisuals != NULL)
  {
    nxagentVisuals = (XVisualInfo *) realloc(nxagentVisuals,
                                                 nxagentNumVisuals * sizeof(XVisualInfo));
  }

  XFree(viList);

  if (nxagentNumVisuals == 0 || nxagentVisuals == NULL)
    FatalError("Unable to find any visuals.\n");

  if (nxagentUserDefaultClass || nxagentUserDefaultDepth) {
    nxagentDefaultVisualIndex = UNDEFINED;
    for (i = 0; i < nxagentNumVisuals; i++)
      if ((!nxagentUserDefaultClass ||
           nxagentVisuals[i].class == nxagentDefaultClass)
          &&
          (!nxagentUserDefaultDepth ||
           nxagentVisuals[i].depth == nxagentDefaultDepth)) {
        nxagentDefaultVisualIndex = i;
        break;
      }
    if (nxagentDefaultVisualIndex == UNDEFINED)
      FatalError("Unable to find desired default visual.\n");
  }
  else {
    vi.visualid = XVisualIDFromVisual(DefaultVisual(nxagentDisplay,
                                      DefaultScreen(nxagentDisplay)));
    nxagentDefaultVisualIndex = 0;
    for (i = 0; i < nxagentNumVisuals; i++)
      if (vi.visualid == nxagentVisuals[i].visualid)
        nxagentDefaultVisualIndex = i;
  }

  nxagentNumDefaultColormaps = nxagentNumVisuals;
  nxagentDefaultColormaps = (Colormap *)xalloc(nxagentNumDefaultColormaps *
                                             sizeof(Colormap));
  for (i = 0; i < nxagentNumDefaultColormaps; i++)
    nxagentDefaultColormaps[i] = XCreateColormap(nxagentDisplay,
                                               DefaultRootWindow(nxagentDisplay),
                                               nxagentVisuals[i].visual,
                                               AllocNone);

  #ifdef WATCH

  fprintf(stderr, "nxagentOpenDisplay: Watchpoint 4.\n");

/*
Reply   Total	Cached	Bits In			Bits Out		Bits/Reply	  Ratio
------- -----	------	-------			--------		----------	  -----
N/A
*/

  sleep(30);

  #endif

  nxagentBlackPixel = BlackPixel(nxagentDisplay, DefaultScreen(nxagentDisplay));
  nxagentWhitePixel = WhitePixel(nxagentDisplay, DefaultScreen(nxagentDisplay));

  #ifdef WATCH

  fprintf(stderr, "nxagentOpenDisplay: Watchpoint 5.\n");

/*
Reply   Total	Cached	Bits In			Bits Out		Bits/Reply	  Ratio
------- -----	------	-------			--------		----------	  -----
N/A
*/

  sleep(30);

  #endif

  /*
   * Initialize the agent's event mask that will be requested
   * for the root and all the top level windows. If the nested
   * window is a child of an existing window, we will need to
   * receive StructureNotify events. If we are going to manage
   * the changes in root window's visibility we'll also need
   * VisibilityChange events.
   */

/*
FIXME: Use of nxagentParentWindow is strongly deprecated.
       We need also to clarify which events are selected
       in the diferent operating modes.
*/

  nxagentInitDefaultEventMask();

  /*
   * Get the pixmap depths and format supported
   * by the real display.
   */

  nxagentGetDepthsAndPixmapFormats();

  /*
   * Create a pixmap for each depth matching the
   * local supported formats with format available
   * on the remote display.
   */

  nxagentSetDefaultDrawables();

  #ifdef RENDER
  if (nxagentRenderEnable)
  {
    nxagentRenderExtensionInit();
  }
  #endif

  /*
   * This GC is referenced in Cursor.c. It can be
   * probably removed.
   */

  nxagentBitmapGC = XCreateGC(nxagentDisplay, nxagentDefaultDrawables[1], 0L, NULL);

  /*
   * Note that this "confine window" is useless at the
   * moment as we reimplement nxagentConstrainCursor()
   * to skip the "constrain" stuff.
   */

  #ifdef TEST
  fprintf(stderr, "nxagentOpenDisplay: Going to create agent's confine window.\n");
  #endif

  nxagentConfineWindow = XCreateWindow(nxagentDisplay,
                                       DefaultRootWindow(nxagentDisplay),
                                       0, 0, 1, 1, 0, 0,
                                       InputOnly,
                                       CopyFromParent,
                                       0L, NULL);

  #ifdef TEST
  fprintf(stderr, "nxagentOpenDisplay: Created agent's confine window with id [%ld].\n",
              nxagentConfineWindow);
  #endif

  if (!(nxagentUserGeometry.flag & XValue))
  {
    nxagentChangeOption(RootX, 0);
  }

  if (!(nxagentUserGeometry.flag & YValue))
  {
    nxagentChangeOption(RootY, 0);
  }

  if (nxagentParentWindow == 0)
  {
    if (!(nxagentUserGeometry.flag & WidthValue))
    {
      if (nxagentOption(Fullscreen))
      {
        nxagentChangeOption(RootWidth, DisplayWidth(nxagentDisplay, DefaultScreen(nxagentDisplay)));
      }
      else
      {
        nxagentChangeOption(RootWidth, 3 * DisplayWidth(nxagentDisplay,
                                           DefaultScreen(nxagentDisplay)) / 4);
      }
    }

    if (!(nxagentUserGeometry.flag & HeightValue))
    {
      if (nxagentOption(Fullscreen))
      {
        nxagentChangeOption(RootHeight, DisplayHeight(nxagentDisplay, DefaultScreen(nxagentDisplay)));
      }
      else
      {
        nxagentChangeOption(RootHeight, 3 * DisplayHeight(nxagentDisplay,
                                            DefaultScreen(nxagentDisplay)) / 4);
      }
    }
  }

  if (!nxagentUserBorderWidth)
  {
    nxagentChangeOption(BorderWidth, 1);
  }

  nxagentLogoDepth = DefaultDepth(nxagentDisplay,
                               DefaultScreen(nxagentDisplay)
                              );

  pV = nxagentVisuals[nxagentDefaultVisualIndex];

  r = pV.red_mask;
  g = pV.green_mask;
  b = pV.blue_mask;
  if(!pV.red_mask|| !pV.green_mask|| !pV.blue_mask){
      nxagentLogoBlack = 0x000000;
      nxagentLogoRed   = 0xff0000;
      nxagentLogoWhite = 0xffffff;
  }
  else
  {
      for(or=0,off=0x800000;(r&(off>>or)) == 0; or++);
      for(og=0,off=0x800000;(g&(off>>og)) == 0; og++);
      for(ob=0,off=0x800000;(b&(off>>ob)) == 0; ob++);
      nxagentLogoRed   = nxagentLogoColor(0xff0000);
      nxagentLogoBlack = nxagentLogoColor(0x000000);
      nxagentLogoWhite = 0xffffff;
  }

#ifdef NXAGENT_LOGO_DEBUG
  fprintf(stderr, "mask r: %x g: %x b: %x\n", r, g, b);
  fprintf(stderr, "offset r: %d g: %d b: %d\n", or, og, ob);
#endif

  #ifdef WATCH

  fprintf(stderr, "nxagentOpenDisplay: Watchpoint 5.1.\n");

/*
Reply   Total	Cached	Bits In			Bits Out		Bits/Reply	  Ratio
------- -----	------	-------			--------		----------	  -----
N/A
*/

  sleep(30);

  #endif

  useXpmIcon = nxagentMakeIcon(nxagentDisplay, &nxagentIconPixmap, &nxagentIconShape);

  #ifdef WATCH

  fprintf(stderr, "nxagentOpenDisplay: Watchpoint 5.2.\n");

/*
Reply   Total	Cached	Bits In			Bits Out		Bits/Reply	  Ratio
------- -----	------	-------			--------		----------	  -----
#84     2		512 bits (0 KB) ->	76 bits (0 KB) ->	256/1 -> 38/1	= 6.737:1
*/

  sleep(30);

  #endif

  #ifdef WATCH

  fprintf(stderr, "nxagentOpenDisplay: Watchpoint 6.\n");

/*
Reply   Total	Cached	Bits In			Bits Out		Bits/Reply	  Ratio
------- -----	------	-------			--------		----------	  -----
N/A
*/

  sleep(30);

  #endif

#ifdef NXAGENT_TIMESTAMP
  fprintf(stderr, "Display: nxagentOpenDisplay finished, time is [%d] milliseconds.\n",
          GetTimeInMillis() - startTime);
#endif
  if (nxagentOption(Persistent))
    reconnectDisplayState = EVERYTHING_DONE;
}

void nxagentGetDepthsAndPixmapFormats()
{
  nxagentDepths = XListDepths(nxagentDisplay, DefaultScreen(nxagentDisplay),
                                  &nxagentNumDepths);

  nxagentPixmapFormats = XListPixmapFormats(nxagentDisplay, &nxagentNumPixmapFormats);

  if (nxagentDepths == NULL || nxagentPixmapFormats == NULL)
  {
    #ifdef PANIC
    fprintf(stderr, "nxagentGetDepthsAndPixmapFormats: PANIC! Failed to get available depths and pixmap formats.\n");
    #endif

    FatalError("Error: Failed to get available depths and pixmap formats.");
  }

  #ifdef TEST
  fprintf(stderr, "nxagentGetDepthsAndPixmapFormats: Got [%d] depths and [%d] available pixmap formats.\n",
              nxagentNumDepths, nxagentNumPixmapFormats);
  #endif
}

void nxagentSetDefaultDrawables()
{
  int i, j;

  for (i = 0; i <= MAXDEPTH; i++)
  {
    nxagentDefaultDrawables[i] = None;
  }

  for (i = 0; i < nxagentNumPixmapFormats; i++)
  {
    #ifdef TEST
    fprintf(stderr, "nxagentSetDefaultDrawables: Checking remote pixmap format [%d] with depth [%d] "
                "bits per pixel [%d] scanline pad [%d].\n", i, nxagentPixmapFormats[i].depth,
                    nxagentPixmapFormats[i].bits_per_pixel, nxagentPixmapFormats[i].scanline_pad);
    #endif

    if (nxagentPixmapFormats[i].depth == 24)
    {
      if (nxagentPixmapFormats[i].bits_per_pixel == 24)
      {
        #ifdef TEST
        fprintf(stderr, "nxagentSetDefaultDrawables: WARNING! Assuming remote pixmap "
                    "format [%d] as true 24 bits.\n", i);
        #endif

        nxagentTrue24 = True;
      }
    }

    for (j = 0; j < nxagentNumDepths; j++)
    {
      #ifdef TEST
      fprintf(stderr, "nxagentSetDefaultDrawables: Cheching depth at index [%d] with pixmap depth [%d] "
                  "and display depth [%d].\n", j, nxagentPixmapFormats[i].depth, nxagentDepths[j]);
      #endif

      if ((nxagentPixmapFormats[i].depth == 1 ||
              nxagentPixmapFormats[i].depth == nxagentDepths[j]) &&
                  nxagentDefaultDrawables[nxagentPixmapFormats[i].depth] == None)
      {
        nxagentDefaultDrawables[nxagentPixmapFormats[i].depth] =
            XCreatePixmap(nxagentDisplay, DefaultRootWindow(nxagentDisplay),
                              1, 1, nxagentPixmapFormats[i].depth);

        #ifdef TEST
        fprintf(stderr, "nxagentSetDefaultDrawables: Created default drawable for depth [%d].\n",
                    nxagentPixmapFormats[i].depth);
        #endif
      }
    }

    if (nxagentDefaultDrawables[nxagentPixmapFormats[i].depth] == None)
    {
      #ifdef TEST
      fprintf(stderr, "nxagentSetDefaultDrawables: WARNING! Forcing default drawable for depth [%d].\n",
                  nxagentPixmapFormats[i].depth);
      #endif

      nxagentDefaultDrawables[nxagentPixmapFormats[i].depth] =
          XCreatePixmap(nxagentDisplay, DefaultRootWindow(nxagentDisplay),
                            1, 1, nxagentPixmapFormats[i].depth);
    }
  }
}

void nxagentCloseDisplay()
{
  #ifdef TEST
  fprintf(stderr, "nxagentCloseDisplay: Called with full generation [%d] and display [%p].\n",
              nxagentDoFullGeneration, (void *) nxagentDisplay);
  #endif

  if (nxagentDoFullGeneration == 0 ||
          nxagentDisplay == NULL)
  {
    return;
  }

  /*
   * If nxagentDoFullGeneration is true, all
   * the X resources will be destroyed upon
   * closing the display connection, so there
   * is no real need to generate additional
   * traffic
   */

  xfree(nxagentDefaultColormaps);
  XFree(nxagentVisuals);

  XFree(nxagentDepths);
  XFree(nxagentPixmapFormats);

  nxagentFreeFontCache();
/*
FIXME: Is this needed?

  nxagentFreeFontMatchStuff();
*/

  /*
   * Free the image cache. This is useful
   * for detecting memory leaks.
   */

  if (nxagentDisplay != NULL)
  {
    NXFreeCache(nxagentDisplay);

    NXResetDisplay(nxagentDisplay);
  }

  XCloseDisplay(nxagentDisplay);

  #ifdef TEST
  fprintf(stderr, "nxagentCloseDisplay: Setting the display to NULL.\n");
  #endif

  nxagentDisplay = NULL;
}

Bool nxagentMakeIcon(Display *display, Pixmap *nxIcon, Pixmap *nxMask)
{
  char *env_path = getenv("PATH");
  int lenght_env_path = 0;
  char icon_filename [256];
  char default_path [256];
  char *icon_path = malloc( strlen(env_path) + sizeof(icon_filename) );
  FILE *icon_fp;
  int status;
  Bool existXpmIcon = False;

  Pixmap IconPixmap;
  Pixmap IconShape;

  if (env_path == NULL)
    lenght_env_path = 0;
  else
    lenght_env_path = strlen(env_path) + 1;
  strncpy(icon_filename, "", 255);
  strncpy(default_path, "", 255);

  strcat(icon_filename, NXAGENT_ICON_NAME);
  strcat(default_path,"/usr/NX/share/images/");
  strcat(default_path,icon_filename);

  if ((icon_fp = fopen(default_path, "r")) == NULL)
  {
    char *s;
    char *temp_path = malloc(lenght_env_path + strlen(icon_filename) );
    char *temp_path1 = malloc(lenght_env_path + strlen(icon_filename) );

    strncpy(temp_path, env_path, strlen(env_path));
    strncpy(temp_path1, "", lenght_env_path + strlen(icon_filename) );

    while (strlen(temp_path) > 0)
    {
       s = strpbrk (temp_path, ":");
       if (s == NULL) break;

       strncpy (temp_path1, temp_path , strlen(temp_path) - strlen(s) );
       strncat (temp_path1, "/", 1);
       strncat (temp_path1, icon_filename, strlen(icon_filename));
       if ((icon_fp = fopen(temp_path1, "r")) != NULL)
       {
          fclose (icon_fp);
          existXpmIcon = True;
          strcpy(icon_path,temp_path1);
          break;
       }
       strncpy(temp_path1, "", lenght_env_path + strlen(icon_filename) );
       strncpy(temp_path1, s + 1, strlen(s)-1);
       strncpy(temp_path, "", lenght_env_path + strlen(icon_filename) );
       strcpy(temp_path, temp_path1 );
       strncpy(temp_path1, "", lenght_env_path + strlen(icon_filename) );
     }
     free(temp_path);
     free(temp_path1);
  }
  else
  {
     fclose (icon_fp);
     existXpmIcon = True;
     strcpy(icon_path, default_path);
  }

  if (existXpmIcon)
  {
     status = XpmReadFileToPixmap(display,
                                 DefaultRootWindow(display),
                               icon_path,
                               &IconPixmap,
                               &IconShape,
                               NULL);

     if (status != XpmSuccess) {
        fprintf(stderr, "XpmError: %s\n", XpmGetErrorString(status));
        existXpmIcon = False;
     }
  }

  if (!existXpmIcon)
  {
     status = XpmCreatePixmapFromData(display,
                                        DefaultRootWindow(display),
                                        nxagentIconData,
                                        &IconPixmap,
                                        &IconShape,
                                        NULL);

     if (status != XpmSuccess)
     {
        fprintf(stderr, "XpmError: %s\n", XpmGetErrorString(status));

        existXpmIcon = False;
     }
     else
     {
        existXpmIcon = True;
     }
  }

  free(icon_path);
  *nxIcon = IconPixmap;
  *nxMask = IconShape;
  return existXpmIcon;
}

Bool nxagentXServerGeometryChanged()
{
  return  (WidthOfScreen(DefaultScreenOfDisplay(nxagentDisplay)) !=
               WidthOfScreen(DefaultScreenOfDisplay(nxagentDisplayBackup))) ||
                    (HeightOfScreen(DefaultScreenOfDisplay(nxagentDisplay)) !=
                         HeightOfScreen(DefaultScreenOfDisplay(nxagentDisplayBackup)));
}

void nxagentBackupDisplayInfo(void)
{
  #if defined(NXAGENT_RECONNECT_DEBUG) || defined(NXAGENT_RECONNECT_DEBUG)
  fprintf(stderr, "nxagentBackupDisplayInfo\n");
  #endif

  if (nxagentDisplayInfoSaved)
  {
    return;
  }

  /*
   * Since we need the display structure
   * in order to behave correctly when no X
   * connection is available, we must always
   * have a good display record.
   * It can be discarded only when a new X
   * connection is available, so we store it
   * in order to destroy whenever the recon-
   * nection succed.
   */

  nxagentDisplayBackup = nxagentDisplay;
  nxagentBitmapGCBackup = nxagentBitmapGC;
  nxagentPixmapFormatsRecBackup = nxagentPixmapFormats;
  nxagentNumPixmapFormatsRecBackup = nxagentNumPixmapFormats;
  nxagentDepthsRecBackup = nxagentDepths;
  nxagentNumDepthsRecBackup = nxagentNumDepths;
  nxagentNumDefaultColormapsRecBackup = nxagentNumDefaultColormaps;
  nxagentVisualsRecBackup = nxagentVisuals;
  nxagentNumVisualsRecBackup = nxagentNumVisuals;
  if (nxagentVisualHasBeenIgnored)
  {
    xfree(nxagentVisualHasBeenIgnored);
    nxagentVisualHasBeenIgnored = NULL;
  }
  nxagentVisualHasBeenIgnored = xalloc(nxagentNumVisuals * sizeof(Bool));
  DisplayWidthRecBackup = DisplayWidth(nxagentDisplay, DefaultScreen(nxagentDisplay));
  DisplayHeightRecBackup = DisplayHeight(nxagentDisplay, DefaultScreen(nxagentDisplay));
  nxagentRenderEnableRecBackup = nxagentRenderEnable;

  nxagentDisplayInfoSaved = True;
}

void nxagentCleanupBackupDisplayInfo(void)
{
  #if defined(NXAGENT_RECONNECT_DEBUG) || defined(NXAGENT_RECONNECT_DISPLAY_DEBUG)
  fprintf(stderr, "nxagentCleanupBackupDisplayInfo\n");
  #endif

  xfree(nxagentPixmapFormatsRecBackup);
  nxagentNumPixmapFormatsRecBackup = 0;

  xfree(nxagentDepthsRecBackup);
  nxagentNumDepthsRecBackup = 0;

  nxagentNumDefaultColormapsRecBackup = 0;

  xfree(nxagentVisualsRecBackup);
  nxagentNumVisualsRecBackup = 0;

  if (nxagentVisualHasBeenIgnored)
  {
    xfree(nxagentVisualHasBeenIgnored);
    nxagentVisualHasBeenIgnored = NULL;
  }

  DisplayWidthRecBackup = 0;
  DisplayHeightRecBackup = 0;

  if (nxagentDisplayBackup)
  {
    XCloseDisplay(nxagentDisplayBackup);

    nxagentDisplayBackup = NULL;
  }

  if (nxagentBitmapGCBackup)
  {
    if (nxagentDisplayBackup)
    {
      XFreeGC(nxagentDisplayBackup, nxagentBitmapGCBackup);
    }
    else
    {
      xfree(nxagentBitmapGCBackup);
    }

    nxagentBitmapGCBackup = NULL;
  }

  nxagentDisplayInfoSaved = False;
}

void nxagentDisconnectDisplay(void)
{
  #if defined(NXAGENT_RECONNECT_DEBUG) || defined(NXAGENT_RECONNECT_DISPLAY_DEBUG)
  fprintf(stderr, "nxagentDisconnectDisplay\n");
  #endif

  switch (reconnectDisplayState)
  {
    case EVERYTHING_DONE:
      if (nxagentBitmapGC &&
              nxagentBitmapGCBackup &&
                  (nxagentBitmapGC != nxagentBitmapGCBackup))
      {
        XFreeGC(nxagentDisplay, nxagentBitmapGC);
      }

      nxagentBitmapGC = nxagentBitmapGCBackup;
    case GOT_PIXMAP_FORMAT_LIST:
    case GOT_DEPTH_LIST:
    case ALLOC_DEF_COLORMAP:
      if (nxagentDefaultColormaps)
      {
        int i;

        for (i = 0; i < nxagentNumDefaultColormaps; i++)
        {
          nxagentDefaultColormaps[i] = None;
        }
      }
    case GOT_VISUAL_INFO:
    case OPENED:
      /*
       * Actually we need the nxagentDisplay
       * structure in order to let the agent
       * go when no X connection is available.
       */

      if (nxagentDisplay &&
              nxagentDisplayBackup &&
                  (nxagentDisplay != nxagentDisplayBackup))
      {
        XCloseDisplay(nxagentDisplay);
      }
    case NOTHING:
      nxagentDisplay = nxagentDisplayBackup;
      break;
    default:
      FatalError("nxagentDisconnectDisplay: Display is in unknown state. Can't continue.");
  }

  reconnectDisplayState = NOTHING;
}

static Bool nxagentReloadVisuals(int flexibility)
{
  int i, nitems;
  XVisualInfo *newVisuals, *vi;
  unsigned long mask = VisualAllMask & ~VisualIDMask & ~VisualBitsPerRGBMask;

  /*
   * The original value of mask was  = VisualAllMask & ~VisualIDMask; 
   * it has been changed to avoid a reconnection error when migrating
   * by sessions with 16-bit depth screen. We realized that visual
   * generated by Xorg and Xfree86 at 16-bit depth differs on the
   * bits_per_rgb value, so we avoid checking it. This change
   * reflects also upon the function.
   */

  newVisuals = malloc(sizeof(XVisualInfo) * nxagentNumVisuals);

  for (i = 0; i < nxagentNumVisuals; i++)
  {
    if ((vi = XGetVisualInfo(nxagentDisplay, mask, nxagentVisuals + i, &nitems)) == NULL)
    {
      #ifdef WARNING
      fprintf(stderr, "reloadVisual: failed to match this visual:\n");
      fprintf(stderr, "\tdepth = %d\n", nxagentVisuals[i].depth);
      fprintf(stderr, "\tclass = %d\n", nxagentVisuals[i].class);
      fprintf(stderr, "\tmask = (%ld,%ld,%ld)\n",
                nxagentVisuals[i].red_mask,
                  nxagentVisuals[i].green_mask,
                    nxagentVisuals[i].blue_mask);
      fprintf(stderr, "\tcolormap size = %d\n", nxagentVisuals[i].colormap_size);
      fprintf(stderr, "\tbits_per_rgb = %d\n", nxagentVisuals[i].bits_per_rgb);
      #endif

      if (nxagentVisuals[i].class == DirectColor)
      {
        #ifdef WARNING
        fprintf(stderr, "Warning: ignoring not matched DirectColor visual.\n");
        #endif

        nxagentVisualHasBeenIgnored[i] = TRUE;
        memcpy(newVisuals + i, nxagentVisuals + i, sizeof(XVisualInfo));
      }
      else
      {
        break;
      }
    }
    else
    {
      memcpy(newVisuals + i, vi, sizeof(XVisualInfo));
      nxagentVisualHasBeenIgnored[i] = FALSE;
      XFree(vi);
    }
  }

  if (i == nxagentNumVisuals)
  {
    nxagentVisuals = newVisuals;

    return TRUE;
  }

  free(newVisuals);

  return FALSE;
}

static Bool checkForColormapCompatibility(int flexibility)
{
  return (nxagentNumDefaultColormaps == nxagentNumDefaultColormapsRecBackup);
}

static Bool checkForDepthCompatibility(int flexibility)
{
  int i;

  if (nxagentNumDepths != nxagentNumDepthsRecBackup)
    return False;

  for (i = 0; i < nxagentNumDepths; i++)
    if (nxagentDepths[i] != nxagentDepthsRecBackup[i])
      return False;

  return True;
}

static Bool checkForPixmapFormatCompatibility(int flexibility)
{
  int i;

  if (nxagentNumPixmapFormats != nxagentNumPixmapFormatsRecBackup)
    return False;

  for (i = 0; i < nxagentNumPixmapFormats; i++)
    if ( (nxagentPixmapFormats[i].depth != nxagentPixmapFormatsRecBackup[i].depth) ||
         (nxagentPixmapFormats[i].bits_per_pixel != nxagentPixmapFormatsRecBackup[i].bits_per_pixel) ||
         (nxagentPixmapFormats[i].scanline_pad != nxagentPixmapFormatsRecBackup[i].scanline_pad) )
      return False;

  return True;
}

Bool nxagentReconnectDisplay(void *p0)
{
  XVisualInfo vi;
  int i;
  int flexibility = *(int*)p0;
  int packMethod, packQuality;

  #if defined(NXAGENT_RECONNECT_DEBUG) || defined(NXAGENT_RECONNECT_DISPLAY_DEBUG)
  fprintf(stderr, "nxagentReconnectDisplay\n");
  #endif

  if (reconnectDisplayState)
  {
    fprintf(stderr, "nxagentReconnectDisplay: Trying to reconnect a session "
                "uncleanly disconnected\n");

    return False;
  }

  /*
   * Reset the values to their defaults.
   */

  packMethod = nxagentPackMethod;
  packQuality = nxagentPackQuality;
  nxagentPackMethod = -1;
  nxagentPackQuality = -1;
  nxagentSplitThreshold = -1;
  nxagentTightThreshold = -1;

  nxagentInstallDisplayHandlers();

  nxagentDisplay = nxagentInternalOpenDisplay(nxagentDisplayName);

  nxagentPostInstallDisplayHandlers();

  if (nxagentDisplay == NULL)
  {
    nxagentSetReconnectError(FAILED_RESUME_DISPLAY_ALERT,
                                 "Couldn't open the display.");

    return FALSE;
  }

  nxagentAddXConnection();

  /*
   * Display is now open.
   */

  reconnectDisplayState = OPENED;

#ifdef NXAGENT_TIMESTAMP
  fprintf(stderr, "Display: display on real X opened, time is [%d] milliseconds.\n",
          GetTimeInMillis() - startTime);
#endif

  nxagentUseNXTrans = nxagentPostProcessArgs(nxagentDisplayName, nxagentDisplay,
                                                 DefaultScreenOfDisplay(nxagentDisplay));

  /*
   * Processing the arguments all the timeouts
   * have been set. Now we have to change the
   * screen-saver timeout.
   */

  nxagentSetScreenSaverTime();

  if (nxagentReloadVisuals(flexibility) == FALSE)
  {
    nxagentSetReconnectError(FAILED_RESUME_VISUALS_ALERT, 
                                 "Couldn't restore the required visuals.");

    return FALSE;
  }

  /*
   * Reconnect the visuals.
   */

  reconnectDisplayState = GOT_VISUAL_INFO;

  if (nxagentUserDefaultClass || nxagentUserDefaultDepth)
  {
    nxagentDefaultVisualIndex = UNDEFINED;

    for (i = 0; i < nxagentNumVisuals; i++)
    {
      if ((!nxagentUserDefaultClass ||
           nxagentVisuals[i].class == nxagentDefaultClass) &&
          (!nxagentUserDefaultDepth ||
           nxagentVisuals[i].depth == nxagentDefaultDepth))
      {
        nxagentDefaultVisualIndex = i;

        break;
      }
    }

    if (nxagentDefaultVisualIndex == UNDEFINED)
    {
      FatalError("Unable to find desired default visual.\n");
    }
  }
  else
  {
    vi.visualid = XVisualIDFromVisual(DefaultVisual(nxagentDisplay, DefaultScreen(nxagentDisplay)));

    nxagentDefaultVisualIndex = 0;

    for (i = 0; i < nxagentNumVisuals; i++)
    {
      if (vi.visualid == nxagentVisuals[i].visualid)
      {
        nxagentDefaultVisualIndex = i;
      }
    }
  }

  nxagentInitAlphaVisual();

  /*
   * Reconnect the colormaps.
   */

  nxagentNumDefaultColormaps = nxagentNumVisuals;

  nxagentDefaultColormaps = (Colormap *) xrealloc(nxagentDefaultColormaps,
                                nxagentNumDefaultColormaps * sizeof(Colormap));

  if (nxagentDefaultColormaps == NULL)
  {
    FatalError("xrealloc failed\n");
  }

  reconnectDisplayState = ALLOC_DEF_COLORMAP;

  for (i = 0; i < nxagentNumDefaultColormaps; i++)
  {
    if (nxagentVisualHasBeenIgnored[i])
    {
      nxagentDefaultColormaps[i] = (XID)0;
    }
    else
    {
      nxagentDefaultColormaps[i] = XCreateColormap(nxagentDisplay,
                                                   DefaultRootWindow(nxagentDisplay),
                                                   nxagentVisuals[i].visual,
                                                   AllocNone);
    }
  }

  if (!checkForColormapCompatibility(flexibility))
  {
    nxagentSetReconnectError(FAILED_RESUME_COLORMAPS_ALERT,
                                 "Number of colormaps has changed.");

    return False;
  }

  nxagentDepths = XListDepths(nxagentDisplay, DefaultScreen(nxagentDisplay),
                            &nxagentNumDepths);

  /*
   * Check the display depth.
   */

  reconnectDisplayState = GOT_DEPTH_LIST;

  if (!checkForDepthCompatibility(flexibility))
  {
    nxagentSetReconnectError(FAILED_RESUME_DEPTHS_ALERT, 
                                 "Couldn't restore all the required depths.");

    return False;
  }

  nxagentPixmapFormats = XListPixmapFormats(nxagentDisplay,
                                          &nxagentNumPixmapFormats);

  /*
   * Check if all the required pixmap
   * formats are supported.
   */

  reconnectDisplayState = GOT_PIXMAP_FORMAT_LIST;

  if (!checkForPixmapFormatCompatibility(flexibility))
  {
    nxagentSetReconnectError(FAILED_RESUME_PIXMAPS_ALERT,
                                 "Couldn't restore all the required pixmap formats.");

    return False;
  }

  #ifdef RENDER

  if (nxagentRenderEnable)
  {
    nxagentRenderExtensionInit();
  }

  if (nxagentRenderEnableRecBackup != nxagentRenderEnable)
  {
    nxagentRenderEnable = nxagentRenderEnableRecBackup;

    nxagentSetReconnectError(FAILED_RESUME_RENDER_ALERT,
                                 "Render extension not available or incompatible version.");

    return False;
  }

  #endif

  nxagentBlackPixel = BlackPixel(nxagentDisplay, DefaultScreen(nxagentDisplay));
  nxagentWhitePixel = WhitePixel(nxagentDisplay, DefaultScreen(nxagentDisplay));

  /*
   * Initialize the agent's event mask that will be requested
   * for the root or all the top level windows. If the nested
   * window is a child of an existing window we will need to
   * receive StructureNotify events. If we are going to manage
   * the changes in root window's visibility we'll also need
   * VisibilityChange events.
   */

  nxagentInitDefaultEventMask();

  /*
   * Create a pixmap for each depth matching the
   * local supported formats with format available
   * on the remote display.
   */

  nxagentSetDefaultDrawables();

  nxagentBitmapGC = XCreateGC(nxagentDisplay, nxagentDefaultDrawables[1], 0L, NULL);

  #ifdef TEST
  fprintf(stderr, "nxagentReconnectDisplay: Going to create agent's confine window.\n");
  #endif

  nxagentConfineWindow = XCreateWindow(nxagentDisplay,
                                     DefaultRootWindow(nxagentDisplay),
                                     0, 0, 1, 1, 0, 0,
                                     InputOnly,
                                     CopyFromParent,
                                     0L, NULL);

  #ifdef TEST
  fprintf(stderr, "nxagentReconnectDisplay: Created agent's confine window with id [%ld].\n",
              nxagentConfineWindow);
  #endif

  nxagentLogoDepth = DefaultDepth(nxagentDisplay, DefaultScreen(nxagentDisplay));

  pV = nxagentVisuals[nxagentDefaultVisualIndex];

  r = pV.red_mask;
  g = pV.green_mask;
  b = pV.blue_mask;
  if(!pV.red_mask|| !pV.green_mask|| !pV.blue_mask)
  {
    nxagentLogoBlack = 0x000000;
    nxagentLogoRed   = 0xff0000;
    nxagentLogoWhite = 0xffffff;
  }
  else
  {
    for(or=0,off=0x800000;(r&(off>>or)) == 0; or++);
    for(og=0,off=0x800000;(g&(off>>og)) == 0; og++);
    for(ob=0,off=0x800000;(b&(off>>ob)) == 0; ob++);
    nxagentLogoRed   = nxagentLogoColor(0xff0000);
    nxagentLogoBlack = nxagentLogoColor(0x000000);
    nxagentLogoWhite = 0xffffff;
  }

#ifdef NXAGENT_LOGO_DEBUG
  fprintf(stderr, "mask r: %x g: %x b: %x\n", r, g, b);
  fprintf(stderr, "offset r: %d g: %d b: %d\n", or, og, ob);
#endif

  useXpmIcon = nxagentMakeIcon(nxagentDisplay, &nxagentIconPixmap, &nxagentIconShape);

  /*
   * All went fine. We can continue
   * handling our clients.
   */

  reconnectDisplayState = EVERYTHING_DONE;

  return True;
}

void nxagentAddXConnection()
{
  int fd = XConnectionNumber(nxagentDisplay);

  nxagentXConnectionNumber = fd;

  #ifdef TEST
  fprintf(stderr, "nxagentAddXConnection: Adding the X connection [%d] "
              "to the device set.\n", nxagentXConnectionNumber);
  #endif

  AddEnabledDevice(nxagentXConnectionNumber);
}

void nxagentRemoveXConnection()
{
  #ifdef TEST
  fprintf(stderr, "nxagentRemoveXConnection: Removing the X connection [%d] "
              "from the device set.\n", nxagentXConnectionNumber);
  #endif

  RemoveEnabledDevice(nxagentXConnectionNumber);
}

/*
 * Force an I/O error and wait until the NX trans-
 * port is gone. It must be called before suspend-
 * ing or terminating a session to ensure that the
 * NX transport is terminated first.
 */

void nxagentTerminateDisplay()
{
  if (nxagentDisplay != NULL)
  {
    #ifdef TEST
    fprintf(stderr, "nxagentTerminateDisplay: Going to shutdown the X connection [%d].\n",
                nxagentXConnectionNumber);
    #endif

    NXForceDisplayError(nxagentDisplay);

    XSync(nxagentDisplay, False);
  }

  #ifdef TEST
  fprintf(stderr, "nxagentTerminateDisplay: Going to wait for the NX transport.\n");
  #endif

  NXTransDestroy(NX_FD_ANY);

  #ifdef TEST
  fprintf(stderr, "nxagentTerminateDisplay: The NX transport is not running.\n");
  #endif
}

/*
 * This has not to do with the remote display but
 * with the X server that the agent is impersonating.
 * We have it here to be consistent with the other
 * cleanup procedures which have mainly to do with
 * the Xlib display connection.
 */

void nxagentAbortDisplay()
{
  /*
   * Be sure the X server socket in .X11-unix is
   * deleted otherwise other users may to become
   * unable to run a session on the same display.
   */

  #ifdef TEST
  fprintf(stderr, "nxagentAbortDisplay: Cleaning up the X server sockets.\n");
  #endif

  CloseWellKnownConnections();
}
