/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2001, 2006 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* NXAGENT, NX protocol compression and NX extensions to this software    */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for applicability.       */
/*                                                                        */
/* NX and NoMachine are trademarks of Medialogic S.p.A.                   */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#include "NXpicturestr.h"
#include "NXglyphstr.h"

#include "Render.h"

#include "X.h"
#include "Xproto.h"

#include "render.h"
#include "renderproto.h"

#include "mi.h"
#include "fb.h"
#include "mipict.h"
#include "fbpict.h"
#include "dixstruct.h"

#include "Agent.h"
#include "Drawable.h"
#include "Trap.h"

#define Atom   XlibAtom
#define Pixmap XlibPixmap
#include "../../../../lib/Xrender/Xrenderint.h"
#undef  Atom
#undef  Pixmap
#include "region.h"
#include "extutil.h"

#include "Display.h"
#include "Pixmap.h"
#include "Cursor.h"
#include "Client.h"
#include "Image.h"

#include "NXproto.h"

#define MAX_FORMATS 255

#define NXAGENT_PICTURE_ALWAYS_POINTS_TO_VIRTUAL

/*
 * Set here the required log level.
 */

#define PANIC
#define WARNING
#undef  TEST
#undef  DEBUG

/*
FIXME: Most operations don't seem to produce any visible result
       but still generate tons of traffic.
*/
#undef  SKIP_LOUSY_RENDER_OPERATIONS
#undef  SKIP_REALLY_ALL_LOUSY_RENDER_OPERATIONS

/*
 * Split the trapezoid requests in
 * multiple messages to help the
 * compression.
 */

#define TRAPEZOIDS_PER_REQUEST  4

/*
 * It is possible to synchronize the whole drawable
 * instead of the glyph extents. Sometimes this could
 * optimize the caching or clean some ugly effects on
 * the screen.
 */

#undef  SYNC_WHOLE_GLYPH_DRAWABLE

int nxagentRenderEnable = UNDEFINED;
int nxagentRenderVersionMajor;
int nxagentRenderVersionMinor;

int nxagentPicturePrivateIndex = 0;

static int picturePrivateCount = 0;
static int nxagentNumFormats = 0;

static XRenderPictFormat nxagentArrayFormats[MAX_FORMATS];

XRenderPictFormat *nxagentMatchingFormats(PictFormatPtr pForm);

BoxPtr nxagentGlyphsExtents;

#ifdef DEBUG

static void nxagentPrintFormat(XRenderPictFormat *pFormat);

#endif

/*
 * From NXglyph.c.
 */

extern const CARD8 glyphDepths[];

Bool nxagentFillGlyphSet(GlyphSetPtr);

/*
 * From NXdispatch.c.
 */

extern void BitOrderInvert(unsigned char *data, int nbytes);

/*
 * Other functions defined here.
 */

void nxagentQueryFormats(void);

void nxagentCreateGlyphSet(GlyphSetPtr pGly);

void nxagentAddGlyphs(GlyphSetPtr glyphSet, Glyph *gids, xGlyphInfo *gi,
                          int nglyphs, CARD8 *images, int sizeImages);

int nxagentCursorSaveRenderInfo(ScreenPtr pScreen, CursorPtr pCursor);

void nxagentCursorPostSaveRenderInfo(CursorPtr pCursor, ScreenPtr pScreen,
                                         PicturePtr pPicture, int x, int y);

int nxagentCreatePicture(PicturePtr pPicture, Mask mask);

void nxagentDestroyPicture(PicturePtr pPicture);

int nxagentChangePictureClip(PicturePtr pPicture, int clipType, int nRects,
                                 xRectangle *rects, int xOrigin, int yOrigin);

void nxagentDestroyPictureClip(PicturePtr pPicture);

void nxagentValidatePicture(PicturePtr pPicture, Mask mask);

void nxagentComposite(CARD8 op, PicturePtr pSrc, PicturePtr pMask, PicturePtr pDst,
                          INT16 xSrc, INT16 ySrc, INT16 xMask, INT16 yMask, INT16 xDst,
                              INT16 yDst, CARD16 width, CARD16 height);

void nxagentGlyphs(CARD8 op, PicturePtr pSrc, PicturePtr pDst,
                       PictFormatPtr maskFormat, INT16 xSrc, INT16 ySrc, int nlists,
                           XGlyphElt8 *elts, int sizeID);

void nxagentCompositeRects(CARD8 op, PicturePtr pDst, xRenderColor *color,
                               int nRect, xRectangle *rects);

void nxagentTrapezoids(CARD8 op, PicturePtr pSrc, PicturePtr pDst,
                           PictFormatPtr maskFormat, INT16 xSrc, INT16 ySrc,
                               int ntrap, xTrapezoid *traps);

void nxagentRasterizeTrapezoid(PicturePtr pMask, xTrapezoid *trap,
                                   int x_off, int y_off);

void nxagentTriangles(CARD8 op, PicturePtr pSrc, PicturePtr pDst,
                          PictFormatPtr maskFormat, INT16 xSrc, INT16 ySrc,
                              int ntri, xTriangle *tris);

void nxagentTriStrip(CARD8 op, PicturePtr pSrc, PicturePtr pDst,
                         PictFormatPtr  maskFormat, INT16 xSrc, INT16 ySrc,
                             int npoint, xPointFixed *points);

void nxagentChangePicture(PicturePtr pPicture, Mask mask);

void nxagentTriFan(CARD8 op, PicturePtr pSrc, PicturePtr pDst,
                       PictFormatPtr maskFormat, INT16 xSrc, INT16 ySrc,
                           int npoint, xPointFixed *points);

void nxagentReferenceGlyphSet(GlyphSetPtr glyphSet);

void nxagentFreeGlyphs(GlyphSetPtr glyphSet, CARD32 *gids, int nglyph);

void nxagentFreeGlyphSet(GlyphSetPtr glyphSet);

void nxagentSetPictureTransform(PicturePtr pPicture, pointer transform);

void nxagentSetPictureFilter(PicturePtr pPicture, char *filter, int name_size,
                                 pointer params, int nparams);

void nxagentReconnectGlyphSet(void* p0, XID x1, void *p2);

Bool nxagentReconnectAllPicture(void *);

Bool nxagentDisconnectAllPicture(void);

void nxagentRenderExtensionInit()
{
  int first_event, first_error;
  int major_version, minor_version;

  if (XRenderQueryExtension(nxagentDisplay, &first_event, &first_error))
  {
    XRenderQueryVersion(nxagentDisplay, &major_version, &minor_version);

    /*
     * As the RENDER requests are passed directly to
     * the remote X server this can cause problems if 
     * our RENDER version is different from the version
     * supported by the remote. For this reasos let's
     * advertise to our clients the lowest between the
     + two versions.
     */

    if (major_version > RENDER_MAJOR || 
            (major_version == RENDER_MAJOR &&
                 minor_version > RENDER_MINOR))
    {
      #ifdef TEST
      fprintf(stderr, "nxagentRenderExtensionInit: Using render version [%d.%d] with "
                  "remote version [%d.%d].\n", RENDER_MAJOR, RENDER_MINOR,
                      major_version, minor_version);
      #endif

      nxagentRenderVersionMajor = RENDER_MAJOR;
      nxagentRenderVersionMinor = RENDER_MINOR;
    }
    else if (major_version < RENDER_MAJOR ||
                 (major_version == RENDER_MAJOR &&
                      minor_version < RENDER_MINOR))
    {
      #ifdef WARNING
      fprintf(stderr, "Info: Local render version %d.%d is higher "
                  "than remote version %d.%d.", RENDER_MAJOR, RENDER_MINOR,
                      major_version, minor_version);

      fprintf(stderr, "Info: Lowering the render version reported to clients.\n");
      #endif

      nxagentRenderVersionMajor = major_version;
      nxagentRenderVersionMinor = minor_version;
    }
    else
    {
      #ifdef TEST
      fprintf(stderr, "nxagentRenderExtensionInit: Local render version %d.%d "
                  "matches remote version %d.%d.\n", RENDER_MAJOR, RENDER_MINOR,
                      major_version, minor_version);
      #endif

      nxagentRenderVersionMajor = major_version;
      nxagentRenderVersionMinor = minor_version;
    }
  }
  else
  {
    #ifdef WARNING
    fprintf(stderr, "Warning: Render not available on the remote display.\n");
    #endif

    nxagentRenderEnable = False;
  }
}

int nxagentCursorSaveRenderInfo(ScreenPtr pScreen, CursorPtr pCursor)
{
  pCursor -> devPriv[pScreen -> myNum] = xalloc(sizeof(nxagentPrivCursor));

  if (nxagentCursorPriv(pCursor, pScreen) == NULL)
  {
    FatalError("xalloc failed");
  }

  nxagentCursorUsesRender(pCursor, pScreen) = 1;
  nxagentCursorPicture(pCursor, pScreen) = NULL;

  return 1;
}

void nxagentCursorPostSaveRenderInfo(CursorPtr pCursor, ScreenPtr pScreen,
                                         PicturePtr pPicture, int x, int y)
{
  nxagentCursorPicture(pCursor, pScreen) = pPicture;
  nxagentCursorXOffset(pCursor, pScreen) = x;
  nxagentCursorYOffset(pCursor, pScreen) = y;
}

int nxagentRenderRealizeCursor(ScreenPtr pScreen, CursorPtr pCursor)
{
  int cid;
  int x, y;

  PicturePtr pPicture;

  pPicture = nxagentCursorPicture(pCursor, pScreen);

  pPicture -> refcnt++;

  x = nxagentCursorXOffset(pCursor, pScreen);
  y = nxagentCursorYOffset(pCursor, pScreen);

  /*
   * Set the lossless trap so that the image functions
   * will not try to encode the image using a lossy
   * compression. Drawables should have a quality flag,
   * telling if they were originally encoded with a
   * lossy algorithm. This would allow us to skip the
   * synchronization if the cursor was already encoded
   * with the best quality.
   */

  #ifdef TEST
  fprintf(stderr, "nxagentRenderRealizeCursor: Forcing the synchronization "
              "of the cursor.\n");
  #endif

  nxagentMarkCorruptedRegion(pPicture -> pDrawable, NULL);

  nxagentLosslessTrap = 1;

  nxagentSynchronizeDrawable(pPicture -> pDrawable, 1);

  nxagentLosslessTrap = 0;

  cid = XRenderCreateCursor(nxagentDisplay, nxagentPicture(pPicture), x, y);

  nxagentCursor(pCursor, pScreen) = cid;

  return 1;
}

int nxagentCreatePicture(PicturePtr pPicture, Mask mask)
{
  XRenderPictureAttributes attributes;
  unsigned long            valuemask=0;
  XRenderPictFormat        *pForm;

  Picture id;

  #ifdef DEBUG
  fprintf(stderr, "nxagentCreatePicture: Function called with picture at [%p].\n",
              (void *) pPicture);
  #endif

  if (pPicture == NULL)
  {
    return 0;
  }

  #ifdef DEBUG

  if (pPicture -> pDrawable -> type == DRAWABLE_PIXMAP)
  {
     if (nxagentIsShmPixmap((PixmapPtr)pPicture -> pDrawable))
     {
       fprintf (stderr, "nxagentCreatePicture: Picture uses a shared pixmap.\n");
     }
     else
     {
       fprintf (stderr, "nxagentCreatePicture: Picture uses a plain pixmap.\n");
     }
  }
  else
  {
     fprintf (stderr, "nxagentCreatePicture: Picture uses a window.\n");
  }

  #endif

  if (mask & CPRepeat)
  {
    attributes.repeat = (Bool)pPicture -> repeat;

    valuemask |= CPRepeat;
  }

  if (mask & CPAlphaMap)
  {
    attributes.alpha_map = nxagentPicturePriv(pPicture -> alphaMap) -> picture;

    valuemask |= CPAlphaMap;
  }

  if (mask & CPAlphaXOrigin)
  {
    attributes.alpha_x_origin = pPicture -> alphaOrigin.x;

    valuemask |= CPAlphaXOrigin;
  }

  if (mask & CPAlphaYOrigin)
  {
    attributes.alpha_y_origin = pPicture -> alphaOrigin.y;

    valuemask |= CPAlphaYOrigin;
  }

  if (mask & CPClipXOrigin)
  {
    attributes.clip_x_origin = pPicture -> clipOrigin.x;

    valuemask |= CPClipXOrigin;
  }

  if (mask & CPClipYOrigin)
  {
    attributes.clip_y_origin = pPicture -> clipOrigin.y;

    valuemask |= CPClipYOrigin;
  }

  if (mask & CPGraphicsExposure)
  {
    attributes.graphics_exposures = (Bool)pPicture -> graphicsExposures;

    valuemask |= CPGraphicsExposure;
  }

  if (mask & CPSubwindowMode)
  {
    attributes.subwindow_mode = pPicture -> subWindowMode;

    valuemask |= CPSubwindowMode;
  }

  if (mask & CPClipMask)
  {
    attributes.clip_mask = None;

    valuemask |= CPClipMask;
  }

  if (mask & CPPolyEdge)
  {
    attributes.poly_edge = pPicture -> polyEdge;

    valuemask |=CPPolyEdge;
  }

  if (mask & CPPolyMode)
  {
    attributes.poly_mode = pPicture -> polyMode;

    valuemask |= CPPolyMode;
  }

  if (mask & CPDither)
  {
    attributes.dither = pPicture -> dither;

    valuemask |= CPDither;
  }

  if (mask & CPComponentAlpha)
  {
    attributes.component_alpha = pPicture -> componentAlpha;

    valuemask |= CPComponentAlpha;
  }

  pForm = NULL;

  if (pPicture -> pFormat != NULL)
  {
    pForm = nxagentMatchingFormats(pPicture -> pFormat);

    #ifdef DEBUG

    nxagentPrintFormat(pForm);

    #endif
  }

  if (pForm == NULL)
  {
    fprintf(stderr, "nxagentCreatePicture: WARNING! The requested format was not found.\n");

    return 0;
  }

  id = XRenderCreatePicture(nxagentDisplay,
                            nxagentDrawable(pPicture -> pDrawable),
                            pForm,
                            valuemask,
                            &attributes);

  #ifdef TEST
  fprintf(stderr, "nxagentCreatePicture: Created picture at [%lx] with drawable at [%p].\n",
              id, (void *) nxagentDrawable(pPicture -> pDrawable));
  #endif

  #ifdef DEBUG

  XSync(nxagentDisplay, False);

  #endif

  nxagentPicturePriv(pPicture) -> picture = id;

  if (nxagentAlphaEnabled == 1 && pPicture -> pDrawable->depth == 32 &&
          pPicture -> pFormat -> direct.alpha != 0)
  {
    if (pPicture -> pDrawable -> type == DRAWABLE_PIXMAP)
    {
      nxagentPixmapPriv(nxagentRealPixmap((PixmapPtr) pPicture -> pDrawable)) -> pPicture = pPicture;
    }
    else if (pPicture -> pDrawable -> type == DRAWABLE_WINDOW)
    {
      nxagentWindowPriv((WindowPtr) pPicture -> pDrawable) -> pPicture = pPicture;
    }
  }

  return Success;
}

XRenderPictFormat *nxagentMatchingFormats(PictFormatPtr pFormat)
{
  int i;

  for (i = 0; i < nxagentNumFormats; i++)
  {
    if (pFormat -> type == nxagentArrayFormats[i].type &&
        pFormat -> depth == nxagentArrayFormats[i].depth &&
        pFormat -> direct.red == nxagentArrayFormats[i].direct.red &&
        pFormat -> direct.green == nxagentArrayFormats[i].direct.green &&
        pFormat -> direct.blue == nxagentArrayFormats[i].direct.blue &&
        pFormat -> direct.redMask == nxagentArrayFormats[i].direct.redMask &&
        pFormat -> direct.greenMask == nxagentArrayFormats[i].direct.greenMask &&
        pFormat -> direct.blueMask == nxagentArrayFormats[i].direct.blueMask &&
        pFormat -> direct.alpha == nxagentArrayFormats[i].direct.alpha &&
        pFormat -> direct.alphaMask == nxagentArrayFormats[i].direct.alphaMask)
    {
      return &nxagentArrayFormats[i];
    }
  }

  #ifdef DEBUG
  fprintf(stderr, "nxagentMatchingFormats: The requested format was not found.\n");
  #endif

  return NULL;
}

void nxagentDestroyPicture(PicturePtr pPicture)
{
  #ifdef DEBUG
  fprintf(stderr, "nxagentDestroyPicture\n");
  #endif

  if (pPicture == NULL)
  {
    return;
  }

  XRenderFreePicture(nxagentDisplay,
                     nxagentPicturePriv(pPicture) -> picture);
  
  #ifdef DEBUG

  XSync(nxagentDisplay, False);

  #endif
}

int nxagentChangePictureClip(PicturePtr pPicture, int clipType, int nRects,
                                 xRectangle *rects, int xOrigin, int yOrigin)
{
  #ifdef DEBUG
  fprintf(stderr, "nxagentChangePictureClip\n");
  #endif

  if (!pPicture)
  {
    return 0;
  }

  switch (clipType)
  {
    case CT_PIXMAP:
    {
      #ifdef DEBUG
      fprintf(stderr, "nxagentChangePictureClip: CT_PIXMAP:\n");
      #endif      
  
      /*
       * if(!nRects)
       * {
       *   return 0;
       * }
       */
/*
FIXME: Is this useful or just a waste of bandwidth?

       Apparently useless with QT.
*/
      #ifndef SKIP_LOUSY_RENDER_OPERATIONS

      XRenderSetPictureClipRectangles(nxagentDisplay,
                                      nxagentPicturePriv(pPicture) -> picture,
                                      xOrigin,
                                      yOrigin,
                                      (XRectangle*)rects,
                                      nRects);
      #endif

      #ifdef DEBUG

      XSync(nxagentDisplay, False);

      #endif

      break;
    }   
    case CT_NONE:
    {
      #ifdef DEBUG
      fprintf(stderr, "nxagentChangePictureClip: CT_NONE:\n");
      #endif
/*
FIXME: Is this useful or just a waste of bandwidth?

       Apparently useless with QT.
*/
      #ifndef SKIP_LOUSY_RENDER_OPERATIONS

      XRenderSetPictureClipRectangles(nxagentDisplay,
                                      nxagentPicturePriv(pPicture) -> picture,
                                      xOrigin,
                                      yOrigin,
                                      (XRectangle*)rects,
                                      nRects);
      #endif

      #ifdef DEBUG

      XSync(nxagentDisplay, False);

      #endif

      break;
    }
    case CT_REGION:
    {
      Region     reg;
      XRectangle rectangle;
      int        index;

      #ifdef DEBUG
      fprintf(stderr, "nxagentChangePictureClip: CT_REGION:\n");
      #endif
    
      reg = XCreateRegion();

      for (index = 0; index <= nRects; index++, rects++)
      {
        rectangle.x = rects -> x;
        rectangle.y = rects -> y;
        rectangle.width = rects -> width;
        rectangle.height = rects -> height;

        XUnionRectWithRegion(&rectangle, reg, reg);
      }
/*
FIXME: Is this useful or just a waste of bandwidth?

       Apparently useless with QT.
*/
      #ifndef SKIP_LOUSY_RENDER_OPERATIONS

      XRenderSetPictureClipRegion(nxagentDisplay,
                                  nxagentPicturePriv(pPicture) -> picture,
                                  reg);

      #endif

      #ifdef DEBUG

      XSync(nxagentDisplay, False);

      #endif
  
      XDestroyRegion(reg);

      break;
    }
    default:
    {
      #ifdef DEBUG
      fprintf(stderr, "nxagentChangePictureClip: clipType not found\n");
      #endif

      break;
    }
  }

  return Success;
}

void nxagentDestroyPictureClip(PicturePtr pPicture)
{
  #ifdef DEBUG
  fprintf(stderr, "nxagentDestroyPictureClip\n");
  #endif
}

void nxagentChangePicture(PicturePtr pPicture, Mask mask)
{
  XRenderPictureAttributes  attributes;
  unsigned long             valuemask = 0;

  #ifdef DEBUG
  fprintf(stderr, "nxagentChangePicture\n");
  #endif

  if (pPicture == NULL)
  {
    return;
  }

  if (mask & CPRepeat)
  {
    attributes.repeat = (Bool)pPicture -> repeat;

    valuemask |= CPRepeat;
  }

  if (mask & CPAlphaMap)
  {
    attributes.alpha_map = nxagentPicturePriv(pPicture -> alphaMap) -> picture;

    valuemask |= CPAlphaMap;
  }

  if (mask & CPAlphaXOrigin)
  {
    attributes.alpha_x_origin = pPicture -> alphaOrigin.x;

    valuemask |= CPAlphaXOrigin;
  }

  if (mask & CPAlphaYOrigin)
  {
    attributes.alpha_y_origin = pPicture -> alphaOrigin.y;

    valuemask |= CPAlphaYOrigin;
  }

  if (mask & CPClipXOrigin)
  {
    attributes.clip_x_origin = pPicture -> clipOrigin.x;

    valuemask |= CPClipXOrigin;
  }

  if (mask & CPClipYOrigin)
  {
    attributes.clip_y_origin = pPicture -> clipOrigin.y;

    valuemask |= CPClipYOrigin;
  }

  if (mask & CPGraphicsExposure)
  {
    attributes.graphics_exposures = (Bool)pPicture -> graphicsExposures;

    valuemask |= CPGraphicsExposure;
  }

  if (mask & CPSubwindowMode)
  {
    attributes.subwindow_mode = pPicture -> subWindowMode;

    valuemask |= CPSubwindowMode;
  }

  if (mask & CPClipMask)
  {
    attributes.clip_mask = None;

    valuemask |= CPClipMask;
  }

  if (mask & CPPolyEdge)
  {
    attributes.poly_edge = pPicture -> polyEdge;

    valuemask |=CPPolyEdge;
  }

  if (mask & CPPolyMode)
  {
    attributes.poly_mode = pPicture -> polyMode;

    valuemask |= CPPolyMode;
  }

  if (mask & CPDither)
  {
    attributes.dither = pPicture -> dither;

    valuemask |= CPDither;
  }

  if (mask & CPComponentAlpha)
  {
    attributes.component_alpha = pPicture -> componentAlpha;

    valuemask |= CPComponentAlpha;
  }

  #ifdef TEST

  if (pPicture -> pDrawable -> type == DRAWABLE_PIXMAP)
  {
    fprintf(stderr, "nxagentChangePicture: %sPixmap [%p] Picture [%lx][%p].\n",
                nxagentIsShmPixmap((PixmapPtr)pPicture -> pDrawable) ? "Shared " : "",
                    (void *) pPicture -> pDrawable, nxagentPicturePriv(pPicture) -> picture, (void *) pPicture);
  }

  #endif
/*
FIXME: Is this useful or just a waste of bandwidth?

       Apparently useless with QT.

       Without this the text is not rendered on GTK/Cairo.
*/
  #ifndef SKIP_REALLY_ALL_LOUSY_RENDER_OPERATIONS

  XRenderChangePicture(nxagentDisplay,
                       nxagentPicturePriv(pPicture) -> picture,
                       valuemask,
                       &attributes);
  #endif

  #ifdef DEBUG

  XSync(nxagentDisplay, False);

  #endif
}

void nxagentValidatePicture(PicturePtr pPicture, Mask mask)
{
  #ifdef DEBUG
  fprintf(stderr, "nxagentValidatePicture\n");
  #endif
}

void nxagentComposite(CARD8 op, PicturePtr pSrc, PicturePtr pMask, PicturePtr pDst,
                          INT16 xSrc, INT16 ySrc, INT16 xMask, INT16 yMask, INT16 xDst,
                              INT16 yDst, CARD16 width, CARD16 height)
{
  if (pSrc == NULL || pDst == NULL)
  {
    return;
  }

  #ifdef TEST

  if (pSrc -> pDrawable -> type == DRAWABLE_PIXMAP)
  {
    fprintf(stderr, "nxagentComposite: %sPixmap [%p] source Picture [%lx][%p].\n",
                nxagentIsShmPixmap((PixmapPtr)pSrc -> pDrawable) ? "Shared " : "",
                    (void *) pSrc -> pDrawable, nxagentPicturePriv(pSrc) -> picture, (void *) pSrc);
  }

  if (pDst -> pDrawable -> type == DRAWABLE_PIXMAP)
  {
    fprintf(stderr, "nxagentComposite: %sPixmap [%p] dest Picture [%lx][%p].\n",
                nxagentIsShmPixmap((PixmapPtr)pDst -> pDrawable) ? "Shared " : "",
                    (void *) pDst -> pDrawable, nxagentPicturePriv(pDst) -> picture, (void *) pDst);
  }

  if (pMask && pMask -> pDrawable -> type == DRAWABLE_PIXMAP)
  {
    fprintf(stderr, "nxagentComposite: %sPixmap [%p] Mask Picture [%lx][%p].\n",
                nxagentIsShmPixmap((PixmapPtr)pMask -> pDrawable) ? "Shared " : "",
                    (void *) pMask -> pDrawable, nxagentPicturePriv(pMask) -> picture, (void *) pMask);
  }

  #endif

  /*
   * Synchronize the content of the shared memory pixmap
   * but pay attention at not doing this more than once.
   * We need to wait until the image data has been recom-
   * posed at the X server side or the operation will use
   * the wrong data.
   */
/*
FIXME: What happens if the drawable is already
       being split?
*/
  nxagentSynchronizeShmPixmap(pSrc -> pDrawable, xSrc, ySrc, width, height);

  if (nxagentDrawableStatus(pSrc -> pDrawable) == NotSynchronized)
  {
    #ifdef TEST
    fprintf(stderr, "nxagentComposite: Synchronizing the source drawable [%p].\n",
                (void *) pSrc -> pDrawable);
    #endif

    nxagentSynchronizeDrawable(pSrc -> pDrawable, 1);
  }

  if (pDst -> pDrawable != pSrc -> pDrawable)
  {
    nxagentSynchronizeShmPixmap(pDst -> pDrawable, xDst, yDst, width, height);

    if (nxagentDrawableStatus(pDst -> pDrawable) == NotSynchronized)
    {
      #ifdef TEST
      fprintf(stderr, "nxagentComposite: Synchronizing the destination drawable [%p].\n",
                  (void *) pDst -> pDrawable);
      #endif

      nxagentSynchronizeDrawable(pDst -> pDrawable, 1);
    }
  }

  if (pMask != NULL && pMask -> pDrawable != pSrc -> pDrawable &&
          pMask -> pDrawable != pDst -> pDrawable)
  {
    nxagentSynchronizeShmPixmap(pMask -> pDrawable, xMask, yMask, width, height);

    if (nxagentDrawableStatus(pMask -> pDrawable) == NotSynchronized)
    {
      #ifdef TEST
      fprintf(stderr, "nxagentComposite: Synchronizing the mask drawable [%p].\n",
                  (void *) pMask -> pDrawable);
      #endif

      nxagentSynchronizeDrawable(pMask -> pDrawable, 1);
    }
  }

  XRenderComposite(nxagentDisplay,
                   op,
                   nxagentPicturePriv(pSrc) -> picture,
                   pMask ? nxagentPicturePriv(pMask) -> picture : 0,
                   nxagentPicturePriv(pDst) -> picture,
                   xSrc,
                   ySrc,
                   xMask,
                   yMask,
                   xDst,
                   yDst,
                   width,
                   height);

  #ifdef DEBUG

  XSync(nxagentDisplay, False);

  #endif
}

void nxagentGlyphs(CARD8 op, PicturePtr pSrc, PicturePtr pDst,
                       PictFormatPtr maskFormat, INT16 xSrc, INT16 ySrc, int nlists,
                           XGlyphElt8 *elts, int sizeID)
{
  XRenderPictFormat *pForm;

  #ifdef TEST
  fprintf(stderr, "nxagentGlyphs: Called with source [%p] destination [%p] and size id [%d].\n",
              (void *) pSrc, (void *) pDst, sizeID);
  #endif

  if (pSrc == NULL || pDst == NULL)
  {
    #ifdef WARNING
    fprintf(stderr, "nxagentGlyphs: WARNING! Invalid source [%p] or destination [%p].\n",
                (void *) pSrc, (void *) pDst);
    #endif

    return;
  }

  pForm = NULL;

  if (maskFormat != NULL)
  {
    pForm = nxagentMatchingFormats(maskFormat);
    
    #ifdef DEBUG

    nxagentPrintFormat(pForm);

    #endif

    if (pForm == NULL)
    {
      return;
    }
  }

  /*
   * Need to synchronize the pixmaps involved in
   * the operation before rendering the glyphs
   * on the real X server.
   */

  if (nxagentDrawableStatus(pSrc -> pDrawable) == NotSynchronized)
  {
    #ifdef TEST
    fprintf(stderr, "nxagentGlyphs: Synchronizing source [%s] at [%p].\n",
                pSrc -> pDrawable -> type == DRAWABLE_PIXMAP ? "pixmap" : "window",
                    (void *) pSrc -> pDrawable);
    #endif

    #ifndef SYNC_WHOLE_GLYPH_DRAWABLE

    #ifdef TEST
    fprintf(stderr, "nxagentGlyphs: Synchronizing region [%d,%d,%d,%d] of source [%s] at [%p] "
                "with geometry [%d,%d,%d,%d].\n", nxagentGlyphsExtents -> x1,
                    nxagentGlyphsExtents -> y1, nxagentGlyphsExtents -> x2, nxagentGlyphsExtents -> y2,
                        (pSrc -> pDrawable -> type == DRAWABLE_PIXMAP ? "pixmap" : "window"),
                            (void *) pSrc -> pDrawable, pSrc -> pDrawable -> x, pSrc -> pDrawable -> y,
                                pSrc -> pDrawable -> x + pSrc -> pDrawable -> width,
                                    pSrc -> pDrawable -> y + pSrc -> pDrawable -> height);
    #endif

    nxagentSynchronizeBox(pSrc -> pDrawable, nxagentGlyphsExtents);

    #else

    #ifdef TEST
    fprintf(stderr, "nxagentGlyphs: Synchronizing the whole source [%s] at [%p] "
                "with geometry [%d,%d,%d,%d].\n", 
                    (pSrc -> pDrawable -> type == DRAWABLE_PIXMAP ? "pixmap" : "window"),
                            (void *) pSrc -> pDrawable, pSrc -> pDrawable -> x, pSrc -> pDrawable -> y,
                                pSrc -> pDrawable -> x + pSrc -> pDrawable -> width,
                                    pSrc -> pDrawable -> y + pSrc -> pDrawable -> height);
    #endif

    nxagentSynchronizeBox(pSrc -> pDrawable, NullBox);

    #endif
  }

  if (pSrc -> pDrawable != pDst -> pDrawable &&
          nxagentDrawableStatus(pDst -> pDrawable) == NotSynchronized)
  {
    #ifdef TEST
    fprintf(stderr, "nxagentGlyphs: Synchronizing destination [%s] at [%p].\n",
                pDst -> pDrawable -> type == DRAWABLE_PIXMAP ? "pixmap" : "window",
                    (void *) pDst -> pDrawable);
    #endif

    #ifndef SYNC_WHOLE_GLYPH_DRAWABLE

    #ifdef TEST
    fprintf(stderr, "nxagentGlyphs: Synchronizing region [%d,%d,%d,%d] of destination [%s] at [%p] "
                "with geometry [%d,%d,%d,%d].\n", nxagentGlyphsExtents -> x1,
                    nxagentGlyphsExtents -> y1, nxagentGlyphsExtents -> x2, nxagentGlyphsExtents -> y2,
                        (pDst -> pDrawable -> type == DRAWABLE_PIXMAP ? "pixmap" : "window"),
                            (void *) pDst -> pDrawable, pDst -> pDrawable -> x, pDst -> pDrawable -> y,
                                pDst -> pDrawable -> x + pDst -> pDrawable -> width,
                                    pDst -> pDrawable -> y + pDst -> pDrawable -> height);
    #endif

    nxagentSynchronizeBox(pDst -> pDrawable, nxagentGlyphsExtents);

    #else

    #ifdef TEST
    fprintf(stderr, "nxagentGlyphs: Synchronizing the whole destination [%s] at [%p] "
                "with geometry [%d,%d,%d,%d].\n",
                        (pDst -> pDrawable -> type == DRAWABLE_PIXMAP ? "pixmap" : "window"),
                            (void *) pDst -> pDrawable, pDst -> pDrawable -> x, pDst -> pDrawable -> y,
                                pDst -> pDrawable -> x + pDst -> pDrawable -> width,
                                    pDst -> pDrawable -> y + pDst -> pDrawable -> height);
    #endif

    nxagentSynchronizeBox(pDst -> pDrawable, NullBox);

    #endif

  }

  switch (sizeID)
  {
    case 1:
    {
/*
FIXME: Is this useful or just a waste of bandwidth?

       This is certainly useful but there is a problem with the
       caching. No message seems to be matched by the checksum.
*/
      XRenderCompositeText8(nxagentDisplay,
                              op,
                              nxagentPicturePriv(pSrc)->picture,
                              nxagentPicturePriv(pDst)->picture,
                              pForm,
                              xSrc,
                              ySrc,
                              elts -> xOff,
                              elts -> yOff,
                              (XGlyphElt8*)elts,
                              nlists);

      break;
    }
    case 2:
    {
      XRenderCompositeText16(nxagentDisplay,
                               op,
                               nxagentPicturePriv(pSrc) -> picture,
                               nxagentPicturePriv(pDst) -> picture,
                               pForm,
                               xSrc,
                               ySrc,
                               elts -> xOff,
                               elts -> yOff,
                               (XGlyphElt16*)elts,
                               nlists);

      break;
    }
    case 4:
    {
      XRenderCompositeText32(nxagentDisplay,
                               op,
                               nxagentPicturePriv(pSrc) -> picture,
                               nxagentPicturePriv(pDst) -> picture,
                               pForm,
                               xSrc,
                               ySrc,
                               elts -> xOff,
                               elts -> yOff,
                               (XGlyphElt32*)elts,
                               nlists);

      break;
    }
    default:
    {
      #ifdef WARNING
      fprintf(stderr, "nxagentGlyphs: WARNING! Invalid size id [%d].\n",
                  sizeID);
      #endif

      break;
    }
  }
}

void nxagentCompositeRects(CARD8 op, PicturePtr pDst, xRenderColor *color,
                               int nRect, xRectangle *rects)
{
  #ifdef DEBUG
  fprintf(stderr, "nxagentCompositeRects\n");
  #endif

  if (pDst == NULL)
  {
    return;
  }

  XRenderFillRectangles(nxagentDisplay,
                        op,
                        (Picture)nxagentPicturePriv(pDst) -> picture,
                        (XRenderColor *) color,
                        (XRectangle *) rects,
                        nRect);

  #ifdef DEBUG

  XSync(nxagentDisplay, False);

  #endif
}

void nxagentTrapezoids(CARD8 op, PicturePtr pSrc, PicturePtr pDst,
                           PictFormatPtr maskFormat, INT16 xSrc, INT16 ySrc,
                               int ntrap, xTrapezoid *traps)
{
  XRenderPictFormat *pForm;

  XTrapezoid *current = (XTrapezoid *) traps;

  int remaining = ntrap;

  #ifdef DEBUG
  fprintf(stderr, "nxagentTrapezoids: Source [%p] destination [%p] coordinates "
              "[%d,%d] elements [%d].\n", (void *) pSrc, (void *) pDst,
                  xSrc, ySrc, ntrap);
  #endif

  if (pSrc == NULL || pDst == NULL)
  {
    return;
  }

  pForm = NULL;

  if (maskFormat != NULL)
  {
    pForm = nxagentMatchingFormats(maskFormat);
    
    #ifdef DEBUG

    nxagentPrintFormat(pForm);

    #endif

    if (pForm == NULL)
    {
      return;
    }
  }
/*
FIXME: Is this useful or just a waste of bandwidth?

       Apparently useless with QT.
*/
  #ifndef SKIP_LOUSY_RENDER_OPERATIONS

  #ifdef TEST

  /*
   * This generates big amounts of traffic. We should
   * find a way to skip the requests that do not have
   * effect on the destination or can be realized at
   * later time. The function miTrapezoidBounds() from
   * mitrap.c may be used for the purpose.
   */

  if (pSrc -> pDrawable -> type == DRAWABLE_PIXMAP)
  {
    fprintf(stderr, "nxagentTrapezoids: Source is a pixmap of geometry [%d,%d].\n",
                pSrc -> pDrawable -> width, pSrc -> pDrawable -> height);
  }
  else if (pSrc -> pDrawable -> type == DRAWABLE_WINDOW)
  {
    fprintf(stderr, "nxagentTrapezoids: WARNING! Source is a window of geometry [%d,%d].\n",
                pSrc -> pDrawable -> width, pSrc -> pDrawable -> height);
  }

  if (pDst -> pDrawable -> type == DRAWABLE_PIXMAP)
  {
    fprintf(stderr, "nxagentTrapezoids: Destination is a pixmap of geometry [%d,%d].\n",
                pDst -> pDrawable -> width, pDst -> pDrawable -> height);
  }
  else if (pDst -> pDrawable -> type == DRAWABLE_WINDOW)
  {
    fprintf(stderr, "nxagentTrapezoids: WARNING! Destination is a window of geometry [%d,%d].\n",
                pDst -> pDrawable -> width, pDst -> pDrawable -> height);
  }

  #endif

  while (remaining > 0)
  {
    XRenderCompositeTrapezoids(nxagentDisplay,
                               op,
                               nxagentPicturePriv(pSrc) -> picture,
                               nxagentPicturePriv(pDst) -> picture,
                               pForm,
                               xSrc,
                               ySrc,
                               (XTrapezoid *) current,
                               (remaining > TRAPEZOIDS_PER_REQUEST ?
                                   TRAPEZOIDS_PER_REQUEST : remaining));

    remaining -= TRAPEZOIDS_PER_REQUEST;
    current   += TRAPEZOIDS_PER_REQUEST;
  }

  #endif

  #ifdef DEBUG

  XSync(nxagentDisplay, False);

  #endif
}

void nxagentRasterizeTrapezoid(PicturePtr pMask, xTrapezoid *trap,
                                   int x_off, int y_off)
{
  #ifdef DEBUG
  fprintf(stderr, "nxagentRasterizeTrapezoids\n");
  #endif
}

void nxagentTriangles(CARD8 op, PicturePtr pSrc, PicturePtr pDst,
                          PictFormatPtr maskFormat, INT16 xSrc, INT16 ySrc,
                              int ntri, xTriangle *tris)
{
  XRenderPictFormat *pForm;

  #ifdef DEBUG
  fprintf(stderr, "nxagentTriangles\n");
  #endif

  if (pSrc == NULL || pDst == NULL)
  {
    return;
  }

  pForm = NULL;

  if (maskFormat != NULL)
  {
    pForm = nxagentMatchingFormats(maskFormat);

    #ifdef DEBUG

    nxagentPrintFormat(pForm);

    #endif

    if (pForm == NULL)
    {
      return;
    }
  }

  XRenderCompositeTriangles(nxagentDisplay,
                            op,
                            nxagentPicturePriv(pSrc) -> picture,
                            nxagentPicturePriv(pDst) -> picture,
                            pForm,
                            xSrc,
                            ySrc,
                            (XTriangle*)tris,
                            ntri);

  #ifdef DEBUG

  XSync(nxagentDisplay, False);

  #endif
}

void nxagentTriStrip(CARD8 op, PicturePtr pSrc, PicturePtr pDst,
                         PictFormatPtr  maskFormat, INT16 xSrc, INT16 ySrc,
                             int npoint, xPointFixed *points)
{
  XRenderPictFormat *pForm;

  #ifdef DEBUG
  fprintf(stderr, "nxagentTriStrip\n");
  #endif

  if (pSrc == NULL || pDst == NULL)
  {
    return;
  }

  pForm = NULL;

  if (maskFormat != NULL)
  {
    pForm = nxagentMatchingFormats(maskFormat);

    #ifdef DEBUG

    nxagentPrintFormat(pForm);

    #endif

    if (pForm == NULL)
    {
      return;
    }
  }

  XRenderCompositeTriStrip(nxagentDisplay,
                           op,
                           nxagentPicturePriv(pSrc) -> picture,
                           nxagentPicturePriv(pDst) -> picture,
                           pForm,
                           xSrc,
                           ySrc,
                           (XPointFixed*)points,
                           npoint);

  #ifdef DEBUG

  XSync(nxagentDisplay, False);

  #endif
}

void nxagentTriFan(CARD8 op, PicturePtr pSrc, PicturePtr pDst,
                       PictFormatPtr maskFormat, INT16 xSrc, INT16 ySrc,
                           int npoint, xPointFixed *points)
{
  XRenderPictFormat *pForm;

  #ifdef DEBUG
  fprintf(stderr, "nxagentTriFan\n");
  #endif

  if (pSrc == NULL || pDst == NULL)
  {
    return;
  }

  pForm = NULL;

  if (maskFormat != NULL)
  {
    pForm = nxagentMatchingFormats(maskFormat);

    #ifdef DEBUG

    nxagentPrintFormat(pForm);

    #endif

    if (pForm == NULL)
    {
      return;
    }
  }

  XRenderCompositeTriFan(nxagentDisplay,
                         op,
                         nxagentPicturePriv(pSrc) -> picture,
                         nxagentPicturePriv(pDst) -> picture,
                         pForm,
                         xSrc,
                         ySrc,
                         (XPointFixed*)points,
                         npoint);

  #ifdef DEBUG

  XSync(nxagentDisplay, False);

  #endif
}

#ifndef NXAGENT_UPGRADE

/*
FIXME: In the 3.0.0 port these functions have been moved
       to Picture.c. We can remove them when the port is
       is complete.
*/
int AllocatePicturePrivateIndex()
{
  return picturePrivateCount++;
}

Bool AllocatePicturePrivate(register ScreenPtr pScreen, int index2, unsigned amount)
{
  unsigned oldamount;

  PictureScreenPtr ps = GetPictureScreen(pScreen);

  /*
   * Round up the size for proper alignment.
   */

  amount = ((amount + (sizeof(long) - 1)) / sizeof(long)) * sizeof(long);

  if (index2 >= ps -> PicturePrivateLen)
  {
    unsigned *nsizes = (unsigned *) xrealloc(ps -> PicturePrivateSizes,
                           (index2 + 1) * sizeof(unsigned));
    if (nsizes == 0)
    {
      return 0;
    }

    while (ps -> PicturePrivateLen <= index2)
    {
      nsizes[ps -> PicturePrivateLen++] = 0;

      ps -> totalPictureSize += sizeof(DevUnion);
    }

    ps -> PicturePrivateSizes = nsizes;
  }

  oldamount = ps -> PicturePrivateSizes[index2];

  if (amount > oldamount)
  {
    ps -> PicturePrivateSizes[index2] = amount;

    ps -> totalPictureSize += (amount - oldamount);
  }

  return 1;
}
#endif

void nxagentQueryFormats()
{
  XRenderInfo *xri;
  XExtDisplayInfo *info = NULL;
  XRenderPictFormat *pformat=NULL;

  int i;

  #ifdef DEBUG
  fprintf(stderr, "nxagentQueryFormats\n");
  #endif

  if (XRenderQueryFormats(nxagentDisplay))
  {
    #ifdef DEBUG

    XSync(nxagentDisplay, False);

    #endif

    info = (XExtDisplayInfo *) XRenderFindDisplay(nxagentDisplay);

    #ifdef DEBUG

    XSync(nxagentDisplay, False);

    #endif

    xri = (XRenderInfo *) info -> data;

    pformat = xri -> format;

    for (i = 0; i < xri -> nformat; i++)
    {
      nxagentArrayFormats[i] = *pformat;

      pformat++;
    }

    nxagentNumFormats = i;
  }
}

void nxagentCreateGlyphSet(GlyphSetPtr pGly)
{
  XRenderPictFormat *pForm;

  #ifdef DEBUG
  fprintf(stderr, "nxagentCreateGlyphSet\n");
  #endif

  pForm = NULL;

  if (pGly -> format != NULL)
  {
    pForm = nxagentMatchingFormats(pGly -> format);

    #ifdef DEBUG

    nxagentPrintFormat(pForm);

    #endif

    if (pForm == NULL)
    {
      return;
    }
  }

  pGly -> remoteID = XRenderCreateGlyphSet(nxagentDisplay, pForm);

  #ifdef DEBUG

  XSync(nxagentDisplay, False);

  #endif
}

void nxagentAddGlyphs(GlyphSetPtr glyphSet, Glyph *gids, xGlyphInfo *gi,
                          int nglyphs, CARD8 *images, int sizeImages)
{
  #ifdef DEBUG
  fprintf(stderr, "nxagentAddGlyphs\n");
  #endif

  if (glyphDepths[glyphSet -> fdepth] == 1 &&
          nxagentServerOrder() != BitmapBitOrder(nxagentDisplay))
  {
    BitOrderInvert ((unsigned char *) images, sizeImages);
  }

  XRenderAddGlyphs(nxagentDisplay,
                   glyphSet -> remoteID,
                   gids,
                   (XGlyphInfo*)(gi),
                   nglyphs,
                   (char*)images,
                   sizeImages);

  #ifdef DEBUG

  XSync(nxagentDisplay, False);

  #endif
}

void nxagentReferenceGlyphSet(GlyphSetPtr glyphSet)
{
  XRenderReferenceGlyphSet (nxagentDisplay, glyphSet -> remoteID);
}

void nxagentFreeGlyphs(GlyphSetPtr glyphSet, CARD32 *gids, int nglyph)
{
  /*
   * This function should require a Glyph pointer.
   */

  XRenderFreeGlyphs(nxagentDisplay, glyphSet -> remoteID, (Glyph *) gids, nglyph);
}

void nxagentFreeGlyphSet(GlyphSetPtr glyphSet)
{
  XRenderFreeGlyphSet(nxagentDisplay, glyphSet -> remoteID);
}

void nxagentSetPictureTransform(PicturePtr pPicture, pointer transform)
{
/*
FIXME: Is this useful or just a waste of bandwidth?

       Apparently useless with QT.
*/
  #ifndef SKIP_LOUSY_RENDER_OPERATIONS

  XRenderSetPictureTransform(nxagentDisplay,
                                 nxagentPicturePriv(pPicture) -> picture,
                                     (XTransform *) transform);
  #endif
}

void nxagentSetPictureFilter(PicturePtr pPicture, char *filter, int name_size,
                                 pointer params, int nparams)
{
  char *szFilter = Xmalloc(name_size + 1);

  if (szFilter == NULL)
  {
    fprintf(stderr, "nxagentSetPictureFilter: error allocating memory for filter name.\n");

    return;
  }

  strncpy(szFilter, filter, name_size);

  szFilter[name_size] = 0;

  #ifdef TEST
  fprintf(stderr, "nxagentSetPictureFilter: Going to set the filter [%s].\n", szFilter);
  #endif
/*
FIXME: Is this useful or just a waste of bandwidth?

       Apparently useless with QT.
*/
  #ifndef SKIP_LOUSY_RENDER_OPERATIONS

  XRenderSetPictureFilter(nxagentDisplay,
                          nxagentPicturePriv(pPicture) -> picture,
                          szFilter,
                          (XFixed *) params,
                          nparams);
  #endif

  Xfree(szFilter);
}


Bool nxagentPictureInit(ScreenPtr pScreen, PictFormatPtr formats, int nformats)
{
  #ifdef RENDER

  #ifdef DEBUG
  fprintf(stderr, "nxagentPictureInit\n");
  #endif

  nxagentQueryFormats();

  if (fbPictureInit(pScreen, formats, nformats) == 0)
  {
    return FALSE;
  }

  nxagentPicturePrivateIndex = AllocatePicturePrivateIndex();

  AllocatePicturePrivate(pScreen, nxagentPicturePrivateIndex, sizeof(nxagentPrivPictureRec));

  #endif

  return TRUE;
}

#ifdef DEBUG

static void nxagentPrintFormat(XRenderPictFormat *pFormat)
{
  if (pFormat == NULL)
  {
    fprintf(stderr, "nxagentPrintFormat: WARNING! null pointer passed to function.\n");

    return;
  }

  fprintf(stderr, "nxagentPrintFormat: Dumping information for format at [%p]:\n\
                   type=%d\n\
                   depth=%d\n\
                   red=%d\n\
                   redMask=%d\n\
                   green=%d\n\
                   greenMask=%d\n\
                   blue=%d\n\
                   blueMask=%d\n\
                   alpha=%d\n\
                   alphaMask=%d\n",
                   (void *) pFormat,
                   pFormat -> type,
                   pFormat -> depth,
                   pFormat -> direct.red,
                   pFormat -> direct.redMask,
                   pFormat -> direct.green,
                   pFormat -> direct.greenMask,
                   pFormat -> direct.blue,
                   pFormat -> direct.blueMask,
                   pFormat -> direct.alpha,
                   pFormat -> direct.alphaMask);
}

#endif

void nxagentReconnectGlyphSet(void* p0, XID x1, void *p2)
{
  GlyphSetPtr pGly = (GlyphSetPtr) p0;

  XRenderPictFormat *pForm = NULL;

  #ifdef DEBUG
  fprintf(stderr, "nxagentReconnectGlyphSet: %p\n", (void *) pGly);
  #endif

  if (pGly -> format)
  {
    pForm = nxagentMatchingFormats(pGly -> format);
  }

  pGly -> remoteID = XRenderCreateGlyphSet(nxagentDisplay, pForm);

  #ifdef DEBUG

  XSync(nxagentDisplay, False);

  #endif

  nxagentFillGlyphSet(pGly);
}

void nxagentReconnectPicture(pointer p0, XID x1, void *p2)
{
  PicturePtr pPicture = (PicturePtr) p0;
  Bool *pBool = (Bool *) p2;
  unsigned long mask = 0;

  XRenderPictureAttributes attributes;
  XRenderPictFormat        *pForm;

  #ifdef DEBUG
  fprintf(stderr, "nxagentReconnectPicture: %p\n", (void *) pPicture);
  #endif

  #ifdef TEST
  fprintf(stderr, "nxagentReconnectPicture: Called with bool [%d] and picture at [%p].\n",
              *pBool, (void *) pPicture);

  fprintf(stderr, "nxagentReconnectPicture: Virtual picture is [%ld].\n",
              nxagentPicture(pPicture));
  #endif

  /*
   * Check if a previous operation has failed
   * and that the involved objects are valid.
   */

  if (*pBool == 0 || pPicture == NULL ||
          nxagentPicture(pPicture) != 0)
  {
    return;
  }

  if (pPicture -> repeat)
  {
    mask |= CPRepeat;

    attributes.repeat = (Bool) pPicture -> repeat;
  }

  if (pPicture -> alphaMap)
  {
    if (!nxagentPicture(pPicture -> alphaMap))
    {
      nxagentReconnectPicture(pPicture -> alphaMap, 0, pBool);

      if (!*pBool || !nxagentPicture(pPicture -> alphaMap))
      {
        return;
      }
    }

    attributes.alpha_map = nxagentPicture(pPicture -> alphaMap);
    attributes.alpha_x_origin = pPicture -> alphaOrigin.x;
    attributes.alpha_y_origin = pPicture -> alphaOrigin.y;

    mask |= (CPAlphaMap | CPAlphaXOrigin | CPAlphaYOrigin);
  }

  if (pPicture -> graphicsExposures)
  {
    attributes.graphics_exposures = pPicture -> graphicsExposures;

    mask |= CPGraphicsExposure;
  }

  attributes.subwindow_mode = pPicture -> subWindowMode;

  mask |= CPSubwindowMode;

  attributes.poly_edge = pPicture -> polyEdge;

  mask |= CPPolyEdge;

  attributes.poly_mode = pPicture -> polyMode;

  mask |= CPPolyMode;

  attributes.dither = pPicture -> dither;

  mask |= CPDither;

  attributes.component_alpha = pPicture -> componentAlpha;

  mask |= CPComponentAlpha;

  pForm = NULL;

  if (pPicture -> pFormat)
  {
    pForm = nxagentMatchingFormats(pPicture -> pFormat);

    #ifdef DEBUG

    nxagentPrintFormat(pForm);

    #endif
  }

  if (!pForm)
  {
    *pBool = False;

    return;
  }

  #ifdef TEST
  fprintf(stderr, "nxagentReconnectPicture: Creating picture at [%p] with drawable [%ld].\n",
              (void *) pPicture, nxagentDrawable(pPicture -> pDrawable));

  fprintf(stderr, "nxagentReconnectPicture: Format is at [%p] mask is [%ld] attributes are at [%p].\n",
              (void *) pForm, mask, (void *) &attributes);
  #endif

  nxagentPicture(pPicture) = XRenderCreatePicture(nxagentDisplay,
                                                  nxagentDrawable(pPicture -> pDrawable),
                                                  pForm,
                                                  mask,
                                                  &attributes);

  #ifdef TEST

  XSync(nxagentDisplay, False);

  #endif

  #ifdef TEST
  fprintf(stderr, "nxagentReconnectPicture: Reconnected picture at [%p] with value [%ld].\n",
              (void *) pPicture, nxagentPicture(pPicture));
  #endif

  if (nxagentAlphaEnabled == 1 && pPicture -> pDrawable->depth == 32 &&
          pPicture -> pFormat -> direct.alpha != 0)
  {
    if (pPicture -> pDrawable -> type == DRAWABLE_PIXMAP)
    {
      nxagentPixmapPriv((PixmapPtr) pPicture -> pDrawable) -> pPicture = pPicture;
    }
    else if (pPicture -> pDrawable -> type == DRAWABLE_WINDOW)
    {
      nxagentWindowPriv((WindowPtr) pPicture -> pDrawable) -> pPicture = pPicture;
    }
  }
}

Bool nxagentReconnectAllPicture(void *p)
{
  int i;
  Bool r;

  #ifdef TEST
  fprintf(stderr, "nxagentReconnectAllPicture: Going to recreate all pictures.\n");
  #endif

  for (i = 0, r = 1; i < MAXCLIENTS; i++)
  {
    if (clients[i])
    {
      FindClientResourcesByType(clients[i], PictureType, nxagentReconnectPicture, &r);

      #ifdef WARNING

      if (r == False)
      {
        fprintf(stderr, "nxagentReconnectAllPicture: WARNING! Failed to recreate "
                    "picture for client [%d].\n", i);
      }

      #endif
    }
  }

  return True;
}

void nxagentDisconnectPicture(pointer p0, XID x1, void* p2)
{
  PicturePtr pPicture = (PicturePtr) p0;
  Bool *pBool = (Bool *) p2;

  #ifdef TEST
  fprintf(stderr, "nxagentDisconnectPicture: Called with bool [%d] and picture at [%p].\n",
              *pBool, (void *) pPicture);

  fprintf(stderr, "nxagentDisconnectPicture: Virtual picture is [%ld].\n",
              nxagentPicture(pPicture));
  #endif

  if (!*pBool || !pPicture)
  {
    return;
  }

  #ifdef DEBUG
  fprintf(stderr, "nxagentDisconnectPicture: %p - XID %lx\n",
              (void *) pPicture, nxagentPicture(pPicture));
  #endif

  nxagentPicture(pPicture) = None;
}

Bool nxagentDisconnectAllPicture()
{
  int i;
  Bool r;

  #ifdef DEBUG
  fprintf(stderr, "nxagentDisconnectAllPicture\n");
  #endif

  for (i = 0, r = 1; i < MAXCLIENTS; i++)
  {
    if (clients[i])
    {
      FindClientResourcesByType(clients[i], PictureType, nxagentDisconnectPicture, &r);

      #ifdef WARNING

      if (r == False)
      {
        fprintf(stderr, "nxagentDisconnectAllPicture: WARNING! Failed to disconnect "
                    "picture for client [%d].\n", i);
      }

      #endif
    }
  }

  return True;
}
